#!/usr/bin/env python3

from optparse import OptionParser
from os import path
from random import randrange
from sys import exit, stderr

from azee.util.debug import setDebugFlag, trace, trace_begin, trace_end
from azee.util.filesys import recursive_file_list
from azee.expr import AZexpr
from azee.dbase import addLocalEntry, known_ref, listLocalEntries, local_db_deref
from azee.interpret import InputReader, read_file_entries, read_file_bindings
from azee.template import recursive_node_match, expr_property
from azee.values import AZeeError, Type

## CONSTANTS
cmd_escape_prefix = '\\'
init_bindings_dir = 'data/init-bindings'
xml_score_output_dir = 'output'
match_results_data_set_name = 'MRES'
prev_working_data_set_name = 'PREV'

## GLOBALS
human_prompt = True
write_xml_scores = False
named_data_sets = {}
working_data_set = []
data_sample_size = 3
last_eval_result = None



def input_lines_until_empty():
    data = []
    while line := input():
        data.append(line)
    return data

def print_expr(expr, file=None):
    print(expr.multiLineStr() if human_prompt else expr.pipedStr(), file=file)



# =================== EVALUATING ===================

def evaluate_expression(expr, bindResult=False):
    global last_eval_result
    last_eval_result = expr.evaluate()
    res_hdr = ['=', Type.toStr(last_eval_result.type)]
    src_pos = expr.srcPos()
    if src_pos.file: # not evaluating direct interactive input
        res_hdr.append('[%s]' % src_pos)
    if bindResult:
        i = 0
        while known_ref(new_key := 'res%i' % i):
            i += 1
        addLocalEntry(new_key, AZexpr.fromValue(last_eval_result))
        res_hdr.append(new_key)
    if write_xml_scores and last_eval_result.isType(Type.SCORE) and src_pos.file:
        bn = path.basename(src_pos.file)
        if bn.endswith('.az'): # remove the typical "az" extension
            bn = path.splitext(bn)[0]
        output_file_name = path.join(xml_score_output_dir, '%s-%i.xml' % (bn, src_pos.line))
        with open(output_file_name, 'w') as f_out:
            f_out.write(last_eval_result.value.xml())
        res_hdr.append('> ' + output_file_name)
    print(*res_hdr)
    if last_eval_result.isType(Type.LIST):
        for e in last_eval_result.value:
            print_expr(e)
    else:
        print(str(last_eval_result)) # CAUTION this might only print a summary or a non-parsable result


def source_file_paths(pathlst, bind_results):
    expr_count = 0
    error_count = 0
    file_count = 0
    for fdpath in pathlst:
        for fpath in recursive_file_list(fdpath):
            trace_begin('Sourcing from %s' % fpath, channel='init')
            try:
                reader = InputReader.fromFile(fpath)
                while not reader.endOfContents():
                    expr = reader.readExpr(expectReadAll=False)
                    evaluate_expression(expr, bindResult=bind_results)
                    expr_count += 1
                file_count += 1
            except AZeeError as err:
                file_count += 1
                error_count += 1
                err.stderrError()
            except FileNotFoundError as err:
                print('No such file or directory: ' + err.filename, file=stderr)
            trace_end(channel='init')
    print('%i file(s) opened; %i expression(s) evaluated; %i error(s) encountered.' % (file_count, expr_count, error_count))



# ============= DATA SET MANAGEMENT =============

def get_data(name):
    try:
        return named_data_sets[name]
    except KeyError:
        raise AZeeError('Unknown data set: ' + name)

def list_data(data):
    for expr in data:
        print_expr(expr)

def sample_data(data):
    selidx = []
    while len(selidx) < data_sample_size and len(selidx) < len(data):
        sel = randrange(len(data))
        while sel in selidx:
            sel = (sel + 1) % len(data)
        selidx.append(sel)
    list_data([data[i] for i in selidx])
    print('%i entries shown out of %i.' % (len(selidx), len(data)))

def load_data_from_paths(pathlst):
    data = []
    for fdpath in pathlst:
        for fpath in recursive_file_list(fdpath):
            trace_begin('Sourcing from %s' % fpath, channel='init')
            try:
                read_file_entries(fpath, data.append)
            except FileNotFoundError as err:
                print('No such file or directory: ' + err.filename, file=stderr)
            trace_end(channel='init')
    working_data_set.extend(data)
    print('%i expression(s) loaded; working data set now has %i entries.' % (len(data), len(working_data_set)))

def save_data_set(name, data):
    verb = 'Updated' if name in named_data_sets else 'Created'
    named_data_sets[name] = data[:]
    print('%s data set "%s" (%d entries).' % (verb, name, len(data)))

def digest_match_results(matches):
    print('%i match(es) found.' % len(matches))
    save_data_set(match_results_data_set_name, [m.matchedExpr() for m in matches])
    if len(matches) > 0:
        for wckey in matches[0].wildcards():
            save_data_set(wckey, [m.matchedWildcard(wckey) for m in matches])




# =================== COMMANDS ===================

def exec_list_data_sets(argstr):
    if argstr:
        print('Unexpected command argument.', file=stderr)
    else:
        for key, data in sorted(named_data_sets.items()):
            print(key, '\t', '%i expressions' % len(data))
        print('Working data set contains %i expression(s).' % len(working_data_set))

def exec_list_expressions(argstr):
    list_data(get_data(argstr) if argstr else working_data_set)

def exec_sample_data(argstr):
    sample_data(get_data(argstr) if argstr else working_data_set)

def exec_sample_size(argstr):
    global data_sample_size
    if argstr:
        try:
            data_sample_size = int(argstr)
            print('Sample size set to %i.' % data_sample_size)
        except ValueError:
            print('Invalid size argument.', file=stderr)
    else:
        print('Sample size:', data_sample_size)

def exec_load_data(argstr):
    if argstr:
        load_data_from_paths([argstr])
    else:
        data = []
        reader = InputReader(input_lines_until_empty())
        while not reader.endOfContents():
            data.append(reader.readExpr(expectReadAll=False))
        working_data_set.extend(data)
        print('%i expression(s) loaded; working data set now has %i entries.' % (len(data), len(working_data_set)))

def exec_use_data(argstr):
    if argstr:
        new_data = get_data(argstr)[:]
        save_data_set(prev_working_data_set_name, working_data_set)
        working_data_set[:] = new_data # this replaces whole list without assigning a "global" variable
        print('Working data set now contains %i expressions.' % len(working_data_set))
    else:
        print('Missing data set name to use.', file=stderr)

def exec_name_data(argstr):
    if argstr:
        if argstr == prev_working_data_set_name or argstr == match_results_data_set_name:
            AZeeError('Using reserved data set name.').stderrWarning()
        save_data_set(argstr, working_data_set)
    else:
        print('Missing argument name for data set.', file=stderr)

def exec_clear_data(argstr):
    if argstr:
        try:
            del named_data_sets[argstr]
            print('Data set deleted.')
        except KeyError:
            print('Unkown data set.', file=stderr)
    else:
        working_data_set.clear()
        print('Working data set cleared.')

def exec_save_to_disk(argstr):
    with open(argstr, 'w') as f:
        for expr in working_data_set:
            pos = expr.srcPos()
            print(file=f)
            print('' if pos is None else '%% [%s]' % pos, file=f)
            print_expr(expr, file=f)
    print('Wrote working data set to file.')

def exec_template_test(argstr):
    if argstr:
        print('Unexpected command argument.', file=stderr)
    else:
        reader = InputReader(input_lines_until_empty())
        template_expr = reader.readExpr(expectReadAll=False)
        tested_expr = reader.readExpr(expectReadAll=True)
        if match := template_expr.templateMatch(tested_expr):
            print('true')
            for wckey in match.wildcards():
                print('*' + wckey)
                print_expr(match.matchedWildcard(wckey))
        else:
            print('false')

def exec_match_expressions(argstr):
    template = InputReader(input_lines_until_empty()).readExpr(expectReadAll=True)
    matched_results = []
    for expr in get_data(argstr) if argstr else working_data_set:
        if match := template.templateMatch(expr):
            matched_results.append(match)
    digest_match_results(matched_results)

def exec_match_nodes(argstr):
    template = InputReader(input_lines_until_empty()).readExpr(expectReadAll=True)
    matched_results = []
    for expr in get_data(argstr) if argstr else working_data_set:
        matched_results.extend(recursive_node_match(template, expr))
    digest_match_results(matched_results)

def exec_histogram(argstr):
    if not argstr:
        print('Missing property to group expressions by.', file=stderr)
    template = InputReader(input_lines_until_empty()).readExpr(expectReadAll=True)
    histogram = {}
    for expr in working_data_set:
        if template.templateMatch(expr):
            key = expr_property(expr, argstr)
            try:
                histogram[key] += 1
            except KeyError:
                histogram[key] = 1
    print(', '.join('%s:%i' % (key, count) for key, count in sorted(histogram.items())))

def exec_bind(argstr):
    if argstr:
        if lines := input_lines_until_empty():
            addLocalEntry(argstr, InputReader(lines).readExpr(expectReadAll=True))
        else:
            print_expr(local_db_deref(argstr))
    else:
        for key in listLocalEntries():
            print(key)

def exec_help(argstr):
    if argstr:
        try:
            print(command_map[argstr][1])
        except KeyError:
            print('No help available for this command.', file=stderr)
    else:
        print('Available commands (use "\\help" with a command keyword for usage): %s'
                % ' '.join('\\' + cmd for cmd in sorted(command_map)))

def exec_src(argstr):
    if argstr:
        source_file_paths([argstr], True)
    else:
        print('Missing argument file.', file=stderr)

def exec_switch_warnings(argstr):
    if argstr:
        try:
            wt = AZeeError.wrn_txt_keys[argstr]
            try:
                AZeeError.silenced_warnings.remove(wt)
            except KeyError:
                AZeeError.silenced_warnings.add(AZeeError.wrn_txt_keys[argstr])
            print('Warnings of type "%s":' % argstr, ['on', 'off'][wt in AZeeError.silenced_warnings])
        except KeyError:
            print('Invalid warning type.', file=stderr)
    else:
        print(', '.join(key for key, wt in AZeeError.wrn_txt_keys.items() if wt not in AZeeError.silenced_warnings))

def exec_xml_score_output(argstr):
    if last_eval_result is not None and last_eval_result.isType(Type.SCORE):
        if argstr:
            with open(argstr, 'w') as f_out:
                f_out.write(last_eval_result.value.xml())
            print('Wrote ' + argstr)
        else:
            print(last_eval_result.value.xml())
    else:
        print('Last result not a SCORE.', file=stderr)




# =================== MAIN ===================

command_map = {
    # Data sets
    'data': (exec_list_data_sets, 'List available data sets and sizes.'),
    'lse': (exec_list_expressions, 'List all expressions in a data set. With arg: in named data set. Without arg: in working data set.'),
    'sd': (exec_sample_data, 'Sample data. With arg: from named data set. Without arg: from working data set.'),
    'ud': (exec_use_data, '"Use data": save working data set to "%s" and load named data set instead.' % prev_working_data_set_name),
    'name': (exec_name_data, 'Name the working data set.'),
    'clr': (exec_clear_data, 'With arg: forget named data set. Without arg: clear working data set.'),
    'load': (exec_load_data, 'Load data to working data set. With arg: read file contents. Without arg: read subsequent expressions until empty line.'),
    'save': (exec_save_to_disk, 'Save working data set to disk under given file path.'),
    # Templating/searching
    'te': (exec_template_test, 'Template expression test. Requires two subsequent expressions before empty line: (1) template; (2) tested expression.'),
    'me': (exec_match_expressions, 'Match expressions; test root nodes only. With arg: in named data set. Without arg: in working data set. Requires subsequent template.'),
    'mn': (exec_match_nodes, 'Match nodes; deep search. With arg: in named data set. Without arg: in working data set. Requires subsequent template.'),
    'distr': (exec_histogram, 'Show distribution of working data filtered with subsequent template, bucketing by argument property (e.g. "prodcount").'),
    # Other
    'bind': (exec_bind, 'With arg: bind key to subsequent expression, or show current if empty line follows. Without arg: list current binding keys.'),
    'help': (exec_help, 'With arg: print command help message. Without arg: list available commands.'),
    'samplesize': (exec_sample_size, 'With arg: set sample size. Without arg: show currently set size.'),
    'src': (exec_src, 'Source expressions from file/directory and evaluate all contents (recursively enter directories).'),
    'wrn': (exec_switch_warnings, 'With arg: switch warnings of given type (valid types: %s). Without arg: list active warnings.' % ', '.join(AZeeError.wrn_txt_keys.keys())),
    'xml': (exec_xml_score_output, 'Write XML for last evaluated result (must be SCORE). With arg: write to file. Without arg: write to standard output.')
}


def top_level_input():
    if human_prompt:
        print('Enter command or AZexpr to evaluate (use "\\help" for available commands):')
    line = input()
    try:
        if line == '':
            pass
        elif line.startswith(cmd_escape_prefix):
            split1 = line.split(maxsplit=1)
            cmd = '' if len(split1) == 0 else split1[0][len(cmd_escape_prefix):]
            argstr = '' if len(split1) < 2 else split1[1]
            try: # commands may raise AZeeError
                command_map[cmd][0](argstr)
            except KeyError as err:
                print('Unrecognised command: %s' % err, file=stderr)
        else:  # evaluate an AZee expression
            expr = InputReader([line] + input_lines_until_empty()).readExpr(expectReadAll=True)
            evaluate_expression(expr, bindResult=True)
    except AZeeError as err:
        print('!!', err.descr())
    print()


def do_main():
    global human_prompt, write_xml_scores
    #setDebugFlag('init')
    #setDebugFlag('eval')
    #setDebugFlag('eval-deref')
    #setDebugFlag('eval-verbose')
    #setDebugFlag('eval-score')
    #setDebugFlag('template')
    parser = OptionParser(description='AZee expression evaluation')
    parser.add_option('--debug', action='append', dest='debug_flags', default=[],
            metavar='FLAGS', help='set debug flags')
    parser.add_option('--bindings', action='append', dest='init_bindings', default=[init_bindings_dir],
            metavar='PATH', help='add file/dir (recursive) with bindings to load on start')
    parser.add_option('--data', action='append', dest='init_data_sets', default=[],
            metavar='DATA_SET', help='load data set on start')
    parser.add_option('--no-prompt', action='store_true', default=False,
            help='good for IPC/piping: no menus or prompts')
    parser.add_option('--no-warn', action='append', dest='no_warn', default=[],
            metavar='WARNINGS', help='silence listed warning types')
    parser.add_option('--xml-scores', action='store_true', default=False,
            help='output XML files for score results sourced from files')

    options, arguments = parser.parse_args()
    human_prompt = not options.no_prompt  # setting global
    write_xml_scores = options.xml_scores # setting global
    for txtarg in options.debug_flags:
        for flag in txtarg.split(','):
            setDebugFlag(flag)
    for txtarg in options.no_warn:
        for warning_type in txtarg.split(','):
            AZeeError.silenced_warnings.add(AZeeError.wrn_txt_keys[warning_type])

    for fdpath in options.init_bindings:
        for fpath in recursive_file_list(fdpath):
            read_file_bindings(fpath, addLocalEntry)
    if options.init_data_sets:
        load_data_from_paths(options.init_data_sets)

    if arguments: # command-line evaluation; exits without an interactive loop
        source_file_paths(arguments, False)
        exit()
    else: # open interactive loop
        trace('Done initialising; starting interactive loop with %d local bindings.' % len(listLocalEntries()), channel='init')
        try:
            while True:
                top_level_input()
        except EOFError:
            if human_prompt:
                print('exit')

if __name__ == '__main__':
    do_main()
