from azee.util.debug import *

from azee import ops
from azee.values import AZeeError, Type, AZval, TypeSpec
from azee.dbase import known_ref, local_db_deref
from azee.spec import linked_bone, site_table
from azee.template import TemplateMatch, template_test



used_infamous_lexical_fallbacks = set()


class AZexpr:
    """
    Class for input (NON-evaluated) AZee expressions.
    Evaluating with the "eval" method yields an AZval object, i.e. a strongly typed value.
    """
    def __init__(self, op, args, srcPos=None):
        assert op is None or isinstance(op, str) and args is None or isinstance(args, list)
        self._value = None
        self._ref = None
        self._op = op
        self._args = args
        self._srcpos = srcPos
        self._type_spec = TypeSpec(None) # default = the non-filtering type spec matching anything
        self._refs = set()
        if op is not None: # None used only by static construction methods below, which set their type spec correctly; should not be used elsewhere
            self._type_spec, self._refs = ops.arg_and_type_check(op, args, srcPos)

    @staticmethod
    def fromValue(az_val_or_typ, val=None, *, srcPos=None):
        res = AZexpr(None, None, srcPos=srcPos)  # dummy "op" and "args" constructor arguments
        res._value = az_val_or_typ if val is None else AZval(az_val_or_typ, val)
        res._type_spec = TypeSpec.fromValue(res._value)
        return res

    @staticmethod
    def mkRef(identifier, srcPos=None):
        assert identifier != '' # so that AZexpr.reference can be used as an "is reference" bool test
        res = AZexpr(None, None, srcPos=srcPos)  # dummy "op" and "args" constructor arguments
        res._ref = identifier
        if known_ref(identifier):
            res._type_spec = local_db_deref(identifier).typeSpec() # could be TypeSpec(None)
        res._refs = {identifier}
        return res

    def __rshift__(self, wanted):
        if self.isValue():
            try:
                return self.value() >> wanted
            except AZeeError as err:
                err.setSrcPos(self.srcPos())
                raise err
        else: # reference or operation
            raise AZeeError('Expected native %s.' % Type.toStr(wanted), src=self.srcPos())

    def __str__(self):
        res = ''
        if self._refs and not self._ref:
            res += '{deps: %s} ' % ', '.join(self._refs) if self._refs else ''
        if len(lines := self._indented_lines(True, ' ', True)) == 1: # always true for values & ref's
            res += lines[0]
        else:
            res += '%s(%s)' % (self.operator(), ', '.join([str(x) for x in self.args()]))
        return res

    def isValue(self, t=None):
        return self._value is not None and (t is None or t == self._value.type)

    def value(self):
        return self._value

    def reference(self):
        return self._ref

    def refDeps(self):
        return self._refs.copy()

    def isOperator(self, op=None):
        return not self.isValue() and not self.reference() and (op is None or op == self._op)

    def operator(self):
        return self._op

    def arg(self, i=0):
        assert len(self._args) > i
        return self._args[i]

    def args(self):
        return self._args

    def typeSpec(self):
        return self._type_spec

    def srcPos(self):
        return self._srcpos

    def possibleInfamousLexicalFallback(self):
        if self.isOperator('apply') and len(self._args) == 1 and (rule := self.arg().reference()):
            if not known_ref(rule):
                return rule
        return None

    ## Expression evaluation
    def evaluate(self):
        used_infamous_lexical_fallbacks.clear()
        result = self.eval({}, None)
        if deps := [dep for dep in self.refDeps() if not known_ref(dep) and dep not in used_infamous_lexical_fallbacks]: # eval already warns of ILFs
            AZeeError('Unbound reference(s) in expression: %s' % ', '.join(deps), src=self.srcPos()).stderrWarning(AZeeError.UNBOUND_REF_DEPS)
        return result

    def eval(self, bind, posture):
        """
        Evaluates the AZ expression and returns an AZvalue object,
        using the given bindings and skeleton if needed.
        """
        try:
            trace_begin('Evaluating: %s' % self, channel='eval')
            # trace('Posture = %s' % posture, channel='eval-verbose')
            trace('with bindings: %s' % ', '.join(str(b) for b in bind), channel='eval-verbose')
            result = None
            if self.isValue():
                result = self.value()
            elif ref := self.reference():
                try:
                    dereferenced = bind[ref]
                except KeyError:
                    dereferenced = local_db_deref(ref) # may raise AeeError
                result = dereferenced.eval({k: v for k, v in bind.items() if k != ref}, posture)
            elif rule := self.possibleInfamousLexicalFallback():
                AZeeError('Infamous lexical fallback: %s' % rule, src=self.srcPos()).stderrWarning(AZeeError.INFAMOUS_LEXICAL_FALLBACK)
                used_infamous_lexical_fallbacks.add(rule)
                result = AZexpr('emptyscore', [], srcPos=self.srcPos()).eval(bind, posture)
            if result is None:
                try:
                    eval_func = getattr(ops, 'op_' + self.operator())
                except AttributeError:
                    raise AZeeError('Invalid operator: %s' % self.operator())
                else:
                    result = eval_func(self.args(), bind, posture)
            result.setSourceExpr(self.deepderef(bind))
            trace_end('Result: %s' % result, channel='eval')
            return result
        except AZeeError as err:  # add source position if known
            if err.position() is None and self._srcpos is not None: # add location to error
                err.setSrcPos(self._srcpos)
            raise err

    def evalAZBodySite(self, bind, posture):
        """
        Evaluate expression to an AZee body site
        """
        trace('Evaluating body site: %s' % self, channel='eval')
        trace('with bindings: %s' % ', '.join(str(b) for b in bind), channel='eval-verbose')
        if ref := self.reference():
            return (bind[ref] if ref in bind else local_db_deref(ref)).evalAZBodySite(bind, posture)
        elif self.isOperator('site'):
            res = ops.id_side_num_str(site_table, self.args(), bind, posture)
            trace_end('Result: %s' % res, channel='eval')
            return res
        else:
            raise AZeeError('Expected site expression.', src=self.srcPos())

    def deepderef(self, bindings, derefLet=True):
        """
        Returns the expression recursively replacing contained references by their dereference in 'bindings'.
        """
        if derefLet and self.isOperator('let'):
            trace_begin('Collapsing "let" expression', channel='eval-deref')
            updated_bindings = bindings.copy()
            i = 0
            while i < len(self.args()) - 1:
                #updated_bindings[self.arg(i) >> Type.STRING] = self.arg(i + 1).deepderef(bindings) # may mask an outer-level binding
                # line version below makes binding order relevant but avoids nesting "let" expr's:
                updated_bindings[self.arg(i) >> Type.STRING] = self.arg(i + 1).deepderef(updated_bindings)
                i += 2
            assert i == len(self.args()) - 1, 'Found even number of args while dereferencing "let" expression'
            res = self.arg(i).deepderef(updated_bindings)
            trace_end('Result: %s' % res, channel='eval-deref')
            return res

        # trace('Available bindings: %s' % bindings.keys())
        if not any(b in self.refDeps() for b in bindings): # significant optimisation
            trace('Shortcut! (expr depends on none of the given bindings)', channel='eval-deref')
            return self

        trace_begin('Replacing ref\'s to %s in expression: %s' % (list(bindings), self), channel='eval-deref')
        if self.isValue():
            trace_end(channel='eval-deref')
            res = self

        elif key := self.reference():
            try:
                res = bindings[key].deepderef({k: v for k, v in bindings.items() if k != key})
            except KeyError:
                res = self

        elif self.isOperator('let'): # keep the expression as a "let" (derefLet already dealt with as top case)
            args = self.args()
            updated_bindings = bindings.copy()
            updated_args = []
            i = 0
            while i < len(args) - 1:
                updated_args.append(args[i])
                updated_args.append(args[i + 1].deepderef(bindings))
                if args[i] >> Type.STRING in updated_bindings:
                    del updated_bindings[args[i] >> Type.STRING]
                i += 2
            assert i == len(args) - 1, 'Found even number of args while dereferencing "let" expression'
            updated_args.append(args[-1].deepderef(updated_bindings))
            res = AZexpr('let', updated_args, srcPos=self.srcPos())

        elif self.isOperator('listfor') or self.isOperator('posturefor'):
            args = self.args()
            new_values = [val.deepderef(bindings) for val in args[1:-1]] # captures both OLD and regular 3-arg syntax of "for"
            new_expr = args[-1].deepderef({k: v for k, v in bindings.items() if k != args[0] >> Type.STRING})
            res = AZexpr(self.operator(), [args[0]] + new_values + [new_expr], srcPos=self.srcPos())

        elif self.isOperator('azop'):
            labels, idict, edict = ops.labelled_arg_list(self.args()[:-1], allowNoExpr=True)
            new_params = [arg if arg.isValue(t=Type.STRING) else arg.deepderef(bindings) for arg in self.args()[:-1]]
            new_body = self.args()[-1].deepderef({k: e for k, e in bindings.items() if k not in labels})
            res = AZexpr('azop', new_params + [new_body], srcPos=self.srcPos())

        else: # no local variable to bind before going deeper
            res = AZexpr(self.operator(), [arg.deepderef(bindings) for arg in self.args()], srcPos=self.srcPos())

        trace_end('Result of deref: %s' % res, channel='eval-deref')
        return res

    def jointDeps(self, bind):
        if self.isValue(): # CHECKME what about ref's? should this method only be called with everything deref'ed?
            return []
        elif ref := self.reference():
            return (bind[ref] if ref in bind else local_db_deref(ref)).jointDeps(bind)
        elif self.isOperator('dir') or self.isOperator('nrm') or self.isOperator('origin'):
            return [self.arg().eval(bind, posture=None) >> Type.BONE]
        elif self.isOperator('site'):
            return [linked_bone(self.evalAZBodySite(bind, posture=None))]
        elif any(self.isOperator(op) for op in ['fwd', 'bwd', 'rgt', 'lft', 'lat', 'contra']):
            return ['back_2']
        else:
            return [dep for arg in self.args() for dep in arg.jointDeps(bind)]

    # Using expression as template ('?' and '*' prefixes recognised)
    def templateMatch(self, tested_expr):
        """
        return a TemplateMatch object if match is successful, None otherwise.
        """
        if self.isValue():
            if tested_expr.isValue() and self.value() == tested_expr.value():
                return TemplateMatch(self)
        elif ref := self.reference():
            if ref == tested_expr.reference():
                return TemplateMatch(self)
        elif self.operator().startswith('*'):
            wckey = self.operator()[1:]
            return TemplateMatch(tested_expr, {wckey: tested_expr}) if wckey else TemplateMatch(tested_expr)
        elif self.operator().startswith('?'):
            trace_begin('Testing template "%s" on %s' % (self.operator(), tested_expr), channel='template')
            result = template_test(self.operator()[1:], self.args(), tested_expr) # may raise AZeeError
            trace_end('Result: %s' % result, channel='template')
            return result
        else: # is non-value expression used as fixed template
            if tested_expr.isOperator(self.operator()) and len(self.args()) == len(tested_expr.args()):
                result = TemplateMatch(self)
                for mine, theirs in zip(self.args(), tested_expr.args()):
                    if match := mine.templateMatch(theirs):
                        result.extendWith(match)
                    else:
                        return None
                return result

    # str output
    def pipedStr(self, shorthand=True):
        return '|'.join(self._indented_lines(shorthand, ' ', False))

    def multiLineStr(self, shorthand=True, compactPipes=True):
        return '\n'.join(self._indented_lines(shorthand, '\t', compactPipes))

    def _indented_lines(self, shorthand, indent_chars, compact_one_line_args):
        if self.isValue():
            v = self.value()
            if v.type in [Type.STRING, Type.NUM, Type.VECT, Type.POINT]: # str(AZval) is known to be parsable with InputReader
                return [str(v)]
            else: # mark unparsable str(AZval)
                return ['Native %s: %s' % (Type.toStr(v.type), v)]
        elif ref := self.reference():
            return ['^' + ref]
        else:
            header = self.operator()
            nested_args = self.args()[:]
            if shorthand:
                if header == 'apply':
                    if nested_args and nested_args[0].reference():
                        header = ':' + nested_args.pop(0).reference()
                elif header == 'partialapply':
                    if nested_args and nested_args[0].reference():
                        header = '.' + nested_args.pop(0).reference()
                elif header == 'bone':
                    header = '!' + (nested_args.pop(0) >> Type.STRING)
                elif header == 'morph':
                    header = '#' + (nested_args.pop(0) >> Type.STRING)
                elif header == 'site':
                    header = '@' + (nested_args.pop(0) >> Type.STRING)
            if compact_one_line_args:
                if all(len(arg_lines := arg._indented_lines(shorthand, indent_chars, False)) == 1
                        and not arg_lines[0].startswith('\'') for arg in nested_args):
                    return [self.pipedStr(shorthand)]
            res = [header]
            for arg in nested_args:
                res.extend(indent_chars + ln for ln in arg._indented_lines(shorthand, indent_chars, compact_one_line_args))
            return res
