
from os import path

from azee.util.debug import trace_begin, trace_end
from azee.util.geom import vec3
from azee.values import Type, AZeeError
from azee.expr import AZexpr



class SourcePosition:
    def __init__(self, file, line=0, pipedSection=None):
        self.file = file # '' means stdin
        self.line = line
        self.piped_line = pipedSection

    def enterPipe(self, section=0):
        assert self.piped_line is None
        return SourcePosition(self.file, self.line, pipedSection=section)

    def isPiped(self):
        return self.piped_line is not None

    def __add__(self, nlines):
        if self.piped_line is None:
            return SourcePosition(self.file, self.line + nlines)
        else:
            return SourcePosition(self.file, self.line, pipedSection=(self.piped_line + nlines))

    def __str__(self):
        if self.file:
            res = '%s:%d' % (path.basename(self.file), self.line)
            if self.piped_line is not None:
                res += ':%d' % self.piped_line
        else:  # no source file (typically '' for stdin)
            if self.piped_line is None:
                res = 'line %d' % self.line
            else:
                res = 'piped section '
                res += str(self.piped_line) if self.line == 1 else '%d:%d' % (self.line, self.piped_line)
        return res




def _readVec3(string):
    try:
        coords = string.split(',')
        if len(coords) != 3:
            raise ValueError(string)
        return vec3(*[float(coord.strip()) for coord in coords])
    except ValueError:
        raise AZeeError('Bad coordinates: %s' % string)


class InputReader:
    def __init__(self, data_lines, initPos=None):
        self._feed = data_lines # consumed
        self._src_pos = initPos if initPos else SourcePosition('')
        self._next_line = None # resets to None if EOF reached
        self._next_contents = ''
        self._next_indent = ''
        self._last_indent = ''
        self._scanForward()

    @staticmethod
    def fromFile(file_name):
        with open(file_name) as f:
            return InputReader(f.read().splitlines(), initPos=SourcePosition(file_name))

    # Digest input until next content line
    def _scanForward(self):
        self._last_indent = self._next_indent
        new_contents = ''
        while self._feed and new_contents == '':
            self._next_line = self._feed.pop(0)
            self._src_pos += 1
            if '%' in self._next_line:
                line = self._next_line.split('%', maxsplit=1)[0].rstrip() # strip off comment part and preceding whitespace
            else:
                line = self._next_line # CAUTION: may contain trailing whitespace
            indent_idx = next((i for i, c in enumerate(line) if not c.isspace()), len(line))
            self._next_indent = line[:indent_idx]
            new_contents = line[indent_idx:]
            if new_contents and new_contents[-1].isspace():
                AZeeError('Trailing whitespace', src=self._src_pos).stderrWarning(AZeeError.WHITESPACE)
                new_contents = new_contents.rstrip()

        if new_contents:
            lmin = min(len(self._last_indent), len(self._next_indent))
            if self._last_indent[:lmin] != self._next_indent[:lmin]:
                raise AZeeError('Indentation does not match previous tab/space mix', src=self._src_pos)
            self._next_contents = new_contents
        else: # EOF reached
            self._next_line = None

    def endOfContents(self):
        return self._next_line is None

    def nextIndentUnder(self, header_indent):
        return not self.endOfContents() and self._next_indent.startswith(header_indent) and len(self._next_indent) > len(header_indent)

    def readContentLine(self):
        if self._next_line is None:
            raise AZeeError('Unexpected end of input', src=self._src_pos)
        indent = self._next_indent
        contents = self._next_contents
        src_pos = self._src_pos
        self._scanForward()
        return indent, contents, src_pos

    def readExpr(self, warnIfIndented=True, expectReadAll=False):
        header_indent, header, header_pos = self.readContentLine()
        if warnIfIndented and header_indent:
            AZeeError('Unexpected indentation.', src=self._src_pos).stderrWarning(AZeeError.WHITESPACE)

        # CASES WITH NO ARG'S: piped expressions, ref's and atomic values like NUMs
        res = None
        if len(pipe_split := header.split('|')) > 1:
            if idx := next((i for i, ln in enumerate(pipe_split, start=1) if ln == ''), 0):
                raise AZeeError('Empty piped section.', src=(header_pos.enterPipe() + idx))
            res = InputReader([pipe_split[0]] + [' ' + ln for ln in pipe_split[1:]], initPos=header_pos.enterPipe()).readExpr(expectReadAll=True)
        elif header.startswith('^'):
            res = AZexpr.mkRef(header[1:], srcPos=header_pos)
        elif header.startswith('\''):  # string
            res = AZexpr.fromValue(Type.STRING, header[1:], srcPos=header_pos)
        elif header.startswith('[') and header.endswith(']'):
            res = AZexpr.fromValue(Type.POINT, _readVec3(header[1:-1]), srcPos=header_pos)
        elif header.startswith('<') and header.endswith('>'):
            res = AZexpr.fromValue(Type.VECT, _readVec3(header[1:-1]), srcPos=header_pos)
        else:
            try: # see if a NUM can be recognised (keep this test before "partialapply" prefix)
                res = AZexpr.fromValue(Type.NUM, float(header), srcPos=header_pos)
            except ValueError:
                pass
        if res:
            if self.nextIndentUnder(header_indent):
                raise AZeeError('This expression cannot take arguments.', src=header_pos)
        else: # CASES ACCEPTING ARG'S
            args = []
            while self.nextIndentUnder(header_indent):
                args.append(self.readExpr(warnIfIndented=False))

            # AZee shorthand expressions
            if header.startswith('!'):  # Shorthand for a bone
                res = AZexpr('bone', [AZexpr.fromValue(Type.STRING, header[1:], srcPos=header_pos)] + args, srcPos=header_pos)
            elif header.startswith('#'):  # Shorthand for a site
                res = AZexpr('morph', [AZexpr.fromValue(Type.STRING, header[1:], srcPos=header_pos)] + args, srcPos=header_pos)
            elif header.startswith('@'):  # Shorthand for a site
                res = AZexpr('site', [AZexpr.fromValue(Type.STRING, header[1:], srcPos=header_pos)] + args, srcPos=header_pos)
            elif header.startswith(':'):  # Shorthand for "apply ref with args"
                res = AZexpr('apply', [AZexpr.mkRef(header[1:].lstrip(), srcPos=header_pos)] + args, srcPos=header_pos)
            elif header.startswith('.'):  # Shorthand for "partialapply ref with args" (keep this test after NUM constants)
                res = AZexpr('partialapply', [AZexpr.mkRef(header[1:].lstrip(), srcPos=header_pos)] + args, srcPos=header_pos)
            else: # Native AZee operator, including '?' and '*' templating prefixes
                res = AZexpr(header, args, srcPos=header_pos)

        if expectReadAll and not self.endOfContents():
            raise AZeeError('Expected end of input; missing indentation?', src=self._src_pos)
        return res






def read_file_entries(file_name, f_expr):
    reader = InputReader.fromFile(file_name)
    trace_begin('Reading entries from %s...' % file_name, channel='init')
    c = 0
    try:
        while not reader.endOfContents():
            f_expr(reader.readExpr(expectReadAll=False))
            c += 1
    except AZeeError as err:
        err.stderrError()
    trace_end('%d entries read.' % c, channel='init')


def read_file_bindings(file_name, f_binding):
    reader = InputReader.fromFile(file_name)
    trace_begin('Reading bindings from %s...' % file_name, channel='init')
    try:
        while not reader.endOfContents():
            indent, binding_key, src_pos = reader.readContentLine()
            if indent:
                AZeeError('Indented binding key "%s"' % binding_key, src_pos).stderrWarning(AZeeError.WHITESPACE)
            f_binding(binding_key, reader.readExpr(expectReadAll=False))
    except AZeeError as err:
        err.stderrError()
    trace_end(channel='init')
