
import re
from math import pi

from azee.util.geom import vec3, quat, StraightPath, ArcPath, SinePath, CatmullRomPath

from azee.score import SyncedScore, CstrScore, HoldScore
from azee.spec import strong_side, bone_table, morph_table, site_table
from azee.values import AZeeError, AZval, Type, TypeSpec
from azee import expr

ADDR_POINT_CACHE_KEY = 'ADDR' # CAUTION: also used to prompt external avatars
MSSP_POINT_CACHE_KEY = 'MSSP' # CAUTION: also used to prompt external avatars
UP_VECT_CACHE_KEY = 'UP' # CAUTION: also used to prompt external avatars
RGT_VECT_CACHE_KEY = 'RGT'
FWD_VECT_CACHE_KEY = 'FWD'
TINY_DIST_CACHE_KEY = 'TINY'
SMALL_DIST_CACHE_KEY = 'SMALL'
MEDIUM_DIST_CACHE_KEY = 'MEDIUM'
LARGE_DIST_CACHE_KEY = 'LARGE'
HUGE_DIST_CACHE_KEY = 'HUGE'


###########################################
##########     Operator list     ##########
###########################################

# expr = any non-STRING AZexpr
# site = a body site AZexpr, not to resolve to a point
# UPPERCASE = an AZexpr of specified values (if STRING: constant value expected, no eval)
# (no tuples anywhere, parentheses are for grouping only, commas for next arg)

# -------------------------------------------------------------------------------------
#   operator        |             args                    |  ret. type  |  ret. value
# -------------------------------------------------------------------------------------
#
#  let              |  (STRING, expr)*, expr               |  follows last arg; preceding arg's are read and bound sequentially
#  apply            |  AZOP, (STRING, expr)*               |  follows azop's return value
#
#  azop             |  (STRING, expr?)*, expr              |  AZOP
#  partialapply     |  AZOP, (STRING, expr)*               |  AZOP
#
#  bone             |  STRING, SIDE?, NUM?                 |  BONE      |  (usual shorthand prefix: "!")
#
#  place            |  site, POINT                         |  POSTURE   |  single body site placement constraint (IK)
#  orient           |  STRING, BONE, VECT                  |  POSTURE   |  single bone orientation constraint (1st arg: "dir", "nrm" or "dxn" axis)
#  look             |  VECT                                |  POSTURE   |  single eye gaze orientation constraint
#  lookat           |  POINT                               |  POSTURE   |  single eye gaze target constraint
#  morph            |  STRING, SIDE?, NUM                  |  POSTURE   |  single articulation-weight constraint (usual shorthand prefix: "#")
#  trill            |  STRING, BONE                        |  POSTURE   |  single trill (oscillation) constraint around bone axis
#  transpath        |  site, STRING, VECT?                 |  POSTURE   |  control a body site over an interpolation; vect to bend, straight path otherwise
#  posture          |  POSTURE*                            |  POSTURE   |  combine postures
#  posturefor       |  STRING, LIST, POSTURE               |  POSTURE
#
#  emptyscore       |  (none)                              |  SCORE     |  empty score of nominal duration 1
#  posturescore     |  NUM, POSTURE                        |  SCORE     |  score of given duration holding given posture set
#  holdscore        |  (none)                              |  SCORE     |  score block holding another block's contraints in a sync operation
#  sync             |  (STRING, STRING*, (SCORE|POSTURE))+ |  SCORE     |  synchronisation of multiple scores ("blocks")
#  scoreseq         |  NUM, LIST                           |  SCORE     |  time sequence of listed scores, each separated from the previous by given time
#  repeatscore      |  NUM, NUM, SCORE                     |  SCORE     |  int number of times, separation time, score to repeat
#  dynamicpoint     |  STRING, STRING, PATH, SCORE, STRING |  SCORE     |  creates moving point + tracking num, which can be referenced throughout the score
#  movesite         |  NUM, POINT, STRING, PATH            |  SCORE     |  creates a score with an effector (site) following given path and dynamics
#
#  oppside          |  SIDE                                |  SIDE
#  l, r             |  (none)                              |  SIDE
#  s, w             |  (none)                              |  SIDE
#
#  up, dwn          |  (none)                              |  VECT
#  lft, rgt, lat    |  (none)                              |  VECT
#  LFT, RGT, LAT    |  (none)                              |  VECT
#  fwd, bwd, contra |  (none)                              |  VECT
#  FWD, BWD, CONTRA |  (none)                              |  VECT
#  dir, nrm         |  BONE                                |  VECT
#  sidevect         |  SIDE, VECT                          |  VECT      |  opposite vector iff argument side is left
#  rhvect           |  VECT                                |  VECT      |  "right-handed vector", i.e. opposite vector iff strong side is left
#  oppvect          |  VECT                                |  VECT
#  addvect          |  VECT, VECT                          |  VECT
#  scalevect        |  NUM, VECT                           |  VECT
#  crossprod        |  VECT, VECT                          |  VECT
#  vect             |  POINT, POINT                        |  VECT
#  along            |  VECT, NUM                           |  VECT
#  tangentonpath    |  PATH, NUM                           |  VECT
#
#  Mssp             |  (none)                              |  POINT
#  addr             |  (none)                              |  POINT
#  origin           |  BONE                                |  POINT
#  site             |  STRING, SIDE?, NUM?                 |  POINT     |  (usual shorthand prefix: "@")
#  translate        |  POINT, VECT, NUM?                   |  POINT     |  offset point along direction (by distance if given)
#  vtranslate       |  POINT, VECT*                        |  POINT     |  translate point along all following vectors (lengths are relevant)
#  midpoint         |  POINT, POINT                        |  POINT
#  pointonpath      |  PATH, NUM                           |  POINT
#  pointsym         |  POINT, POINT                        |  POINT     |  first arg is centre of symmetry
#  linesym          |  POINT, VECT, POINT                  |  POINT     |  first two arg's define axis of symmetry (direction vector)
#  planesym         |  POINT, VECT, POINT                  |  POINT     |  first two arg's define plane of symmetry (normal vector)
#
#  straightfrom     |  POINT, VECT                         |  PATH
#  straightthrough  |  VECT, POINT                         |  PATH
#  straightto       |  VECT, POINT                         |  PATH
#  straightbetween  |  POINT, POINT                        |  PATH
#  arcfrom          |  POINT, POINT, VECT, NUM             |  PATH      |  starting point, centre point, nrm vect, rev count
#  arcthrough       |  VECT, POINT, VECT, NUM              |  PATH      |  through point trans vect (halfway on path), centre point, nrm vect, rev count
#  arcto            |  POINT, POINT, VECT, NUM             |  PATH      |  end point, centre point, nrm vect, rev count
#  wavyfrom         |  POINT, VECT, VECT, NUM, NUM         |  PATH      |  starting point, vector to end point, nrm vect, amplitude, wave count
#  curve            |  LIST                                |  PATH      |  smooth curve passing through all listed points in order
#  vcurve           |  POINT, VECT*                        |  PATH      |  curve starting at point, each vector gives next point by translation of previous
#
#  opp              |  NUM                                 |  NUM
#  plus             |  NUM, NUM                            |  NUM
#  mult             |  NUM, NUM                            |  NUM
#  div              |  NUM, NUM                            |  NUM
#  small, medium    |  (none)                              |  NUM
#  tiny,large,huge  |  (none)                              |  NUM
#  dist             |  POINT, POINT                        |  NUM
#
#  list             |  expr*                               |  LIST
#  listconcat       |  LIST*                               |  LIST
#  listfor          |  STRING, LIST, expr                  |  LIST
#  repeatlist       |  NUM, expr                           |  LIST


# The following dict contains all operators with fixed (non-variable) and basic (non-AZOP) return types.
# Its values are (argument sig list, return basic type) pairs specifying the signature for the operator key.
# Argument sig elements are:
#  - None: obligatory argument of any type (no type restriction applies);
#  - OPT+t: optional argument of basic type t;
#  - tuple of basic types (should not be empty): obligatory argument of either of the listed types;
#  - list of nested arg signatures (should not be empty): optional and repeatable argument sequence.
OPT = len(Type.aztypes)
operator_signatures = {
	'bone':       ([Type.STRING, OPT+Type.SIDE, OPT+Type.NUM], Type.BONE),

	'posture':    ([[Type.POSTURE]], Type.POSTURE),
	'morph':      ([Type.STRING, OPT+Type.SIDE, Type.NUM], Type.POSTURE),
	'place':      ([Type.POINT, Type.POINT], Type.POSTURE), # first arg must actually resolve to a body site
	'orient':     ([Type.STRING, Type.BONE, Type.VECT], Type.POSTURE),
	'look':       ([Type.VECT], Type.POSTURE),
	'lookat':     ([Type.POINT], Type.POSTURE),
	'trill':      ([Type.STRING, Type.BONE], Type.POSTURE),
	'transpath':  ([Type.POINT, Type.STRING, OPT+Type.VECT], Type.POSTURE), # first arg must actually resolve to a body site
	'posturefor': ([Type.STRING, Type.LIST, Type.POSTURE], Type.POSTURE),

	'emptyscore':   ([], Type.SCORE),
	'posturescore': ([Type.NUM, Type.POSTURE], Type.SCORE),
	'holdscore':    ([], Type.SCORE),
	'sync':         ([Type.STRING, [Type.STRING], (Type.SCORE, Type.POSTURE), [Type.STRING, [Type.STRING], (Type.SCORE, Type.POSTURE)]], Type.SCORE),
	'scoreseq':     ([Type.NUM, Type.LIST], Type.SCORE),
	'repeatscore':  ([Type.NUM, Type.NUM, Type.SCORE], Type.SCORE),
	'dynamicpoint': ([Type.STRING, Type.STRING, Type.PATH, Type.STRING, Type.SCORE], Type.SCORE),
	'movesite':     ([Type.NUM, Type.POINT, Type.STRING, Type.PATH], Type.SCORE), # second arg must actually resolve to a body site

	'oppside': ([Type.SIDE], Type.SIDE),
	'l':       ([], Type.SIDE),
	'r':       ([], Type.SIDE),
	's':       ([], Type.SIDE),
	'w':       ([], Type.SIDE),

	'dir':           ([Type.BONE], Type.VECT),
	'nrm':           ([Type.BONE], Type.VECT),
	'RGT':           ([], Type.VECT),
	'LFT':           ([], Type.VECT),
	'LAT':           ([], Type.VECT),
	'FWD':           ([], Type.VECT),
	'BWD':           ([], Type.VECT),
	'CONTRA':        ([], Type.VECT),
	'rgt':           ([], Type.VECT),
	'fwd':           ([], Type.VECT),
	'up':            ([], Type.VECT),
	'lft':           ([], Type.VECT),
	'bwd':           ([], Type.VECT),
	'dwn':           ([], Type.VECT),
	'lat':           ([], Type.VECT),
	'contra':        ([], Type.VECT),
	'sidevect':      ([Type.SIDE, Type.VECT], Type.VECT),
	'rhvect':        ([Type.VECT], Type.VECT),
	'oppvect':       ([Type.VECT], Type.VECT),
	'addvect':       ([Type.VECT, Type.VECT], Type.VECT),
	'scalevect':     ([Type.NUM, Type.VECT], Type.VECT),
	'crossprod':     ([Type.VECT, Type.VECT], Type.VECT),
	'vect':          ([Type.POINT, Type.POINT], Type.VECT),
	'along':         ([Type.VECT, Type.NUM], Type.VECT),
	'tangentonpath': ([Type.PATH, Type.NUM], Type.VECT),

	'site':        ([Type.STRING, OPT+Type.SIDE, OPT+Type.NUM], Type.POINT),
	'Mssp':        ([], Type.POINT),
	'addr':        ([], Type.POINT),
	'origin':      ([Type.BONE], Type.POINT),
	'translate':   ([Type.POINT, Type.VECT, Type.NUM], Type.POINT),
	'vtranslate':  ([Type.POINT, [Type.VECT]], Type.POINT),
	'midpoint':    ([Type.POINT, Type.POINT], Type.POINT),
	'pointonpath': ([Type.PATH, Type.NUM], Type.POINT),
	'pointsym':    ([Type.POINT, Type.POINT], Type.POINT),
	'linesym':     ([Type.POINT, Type.VECT, Type.POINT], Type.POINT),
	'planesym':    ([Type.POINT, Type.VECT, Type.POINT], Type.POINT),

	'straightfrom':    ([Type.POINT, Type.VECT], Type.PATH),
	'straightthrough': ([Type.VECT, Type.POINT], Type.PATH),
	'straightto':      ([Type.VECT, Type.POINT], Type.PATH),
	'straightbetween': ([Type.POINT, Type.POINT], Type.PATH),
	'arcfrom':         ([Type.POINT, Type.POINT, Type.VECT, Type.NUM], Type.PATH),
	'arcthrough':      ([Type.VECT, Type.POINT, Type.VECT, Type.NUM], Type.PATH),
	'arcto':           ([Type.POINT, Type.POINT, Type.VECT, Type.NUM], Type.PATH),
	'wavyfrom':        ([Type.POINT, Type.VECT, Type.VECT, Type.NUM, Type.NUM], Type.PATH),
	'curve':           ([Type.LIST], Type.PATH),
	'vcurve':          ([Type.POINT, [Type.VECT]], Type.PATH),

	'opp':    ([Type.NUM], Type.NUM),
	'plus':   ([Type.NUM, Type.NUM], Type.NUM),
	'mult':   ([Type.NUM, Type.NUM], Type.NUM),
	'div':    ([Type.NUM, Type.NUM], Type.NUM),
	'small':  ([], Type.NUM),
	'medium': ([], Type.NUM),
	'tiny':   ([], Type.NUM),
	'large':  ([], Type.NUM),
	'huge':   ([], Type.NUM),
	'dist':   ([Type.POINT, Type.POINT], Type.NUM),

	'list':       ([[None]], Type.LIST),
	'listfor':    ([Type.STRING, Type.LIST, None], Type.LIST),
	'listconcat': ([[Type.LIST]], Type.LIST),
	'repeatlist': ([Type.NUM, None], Type.LIST)
}

def ckExprType(tspec, gotexpr):
	gotspec = gotexpr.typeSpec()
	if not tspec.accepts(gotspec):
		raise AZeeError('Expected %s; got %s.' % (TypeSpec.toStr(tspec), TypeSpec.toStr(gotspec)), src=gotexpr.srcPos())

def _usedSignatureArgs(sigspec, exprlst):
	if not sigspec: # nothing more expected
		return 0
	elif isinstance(sigspec[0], list): # optional and repeatable arg type sequence
		try:
			used = _usedSignatureArgs(sigspec[0], exprlst)
			return used + _usedSignatureArgs(sigspec, exprlst[used:])
		except AZeeError:
			return _usedSignatureArgs(sigspec[1:], exprlst)
	elif isinstance(sigspec[0], tuple): # choice of one
		for option in sigspec[0]:
			try:
				return _usedSignatureArgs([option] + sigspec[1:], exprlst)
			except AZeeError:
				pass
		raise AZeeError('Bad number or type of arguments.', src=(exprlst[0] if exprlst else None))
	elif sigspec[0] and sigspec[0] >= OPT: # optional arg
		if exprlst and TypeSpec(sigspec[0] - OPT).accepts(exprlst[0].typeSpec()):
			try:
				return 1 + _usedSignatureArgs(sigspec[1:], exprlst[1:])
			except AZeeError:
				pass
		return _usedSignatureArgs(sigspec[1:], exprlst)
	else: # definitely expected arg type first
		if not exprlst:
			raise AZeeError('Not enough arguments.')
		ckExprType(TypeSpec(sigspec[0]), exprlst[0])
		return 1 + _usedSignatureArgs(sigspec[1:], exprlst[1:])

def ckArgSig(argsig, arglst, src_pos):
	try:
		if _usedSignatureArgs(argsig, arglst) < len(arglst):
			raise AZeeError('Bad number or type of arguments.')
	except AZeeError as err:
		if err.position() is None:
			err.setSrcPos(src_pos)
		raise err


def arg_and_type_check(op, args, src_pos):
	"""
	IMPORTANT NOTE: this does SOME type checking but does not guarantee type error-free evaluations.
	Because of the lazy evaluation system (hanging ref's, etc.), many expressions might still have unresolved types.
	"""
	if op in operator_signatures: # fixed return type (except AZOP)
		arg_sig, ret_sig = operator_signatures[op]
		ckArgSig(arg_sig, args, src_pos)
		if op == 'listfor' or op == 'posturefor':
			ref_deps = args[-1].refDeps()
			ref_deps.discard(args[0] >> Type.STRING)
			ref_deps.update(*(arg.refDeps() for arg in args[1:-1]))
		elif op == 'dynamicpoint': # str str path str score
			pvar, nvar, path, dyn, score = args
			ref_deps = score.refDeps()
			ref_deps.discard(pvar >> Type.STRING)
			ref_deps.discard(nvar >> Type.STRING)
			ref_deps.update(path.refDeps())
		else:
			ref_deps = set().union(*(arg.refDeps() for arg in args))
		return TypeSpec(ret_sig), ref_deps

	# SPECIAL CASES WITH LISTS OF LABELLED ARGUMENTS
	if op == 'azop': # returns AZOP
		if not args:
			raise AZeeError('Expected at least one argument.', src_pos)
		labels, idict, edict = labelled_arg_list(args[:-1], allowNoExpr=True)
		res = TypeSpec(Type.AZOP)
		res._azop_args = {}
		for lbl in labels:
			expr = edict.get(lbl, None)
			try:
				tspec = TypeSpec.fromStr(idict[lbl])
				if expr:
					ckExprType(tspec, expr)
			except KeyError: # no type info after label
				tspec = TypeSpec(None)
			res._azop_args[lbl] = expr is not None, tspec
		res._azop_res = args[-1].typeSpec()
		ref_deps = args[-1].refDeps()
		ref_deps.difference_update(set(labels))
		ref_deps.update(*(e.refDeps() for e in edict.values()))
		return res, ref_deps

	elif op == 'partialapply': # returns AZOP
		if not args:
			raise AZeeError('Expected at least one argument.', src_pos)
		ckExprType(TypeSpec(Type.AZOP), args[0])
		labels, idict, edict = labelled_arg_list(args[1:], allowNoExpr=True)
		if bad := next((lbl for lbl in labels if (lbl in idict) == (lbl in edict)), None):
			raise AZeeError('Argument "%s" must either be reduced or renamed.' % bad, src_pos)
		res = TypeSpec(Type.AZOP) # to be filled with arg's & result type specs if able
		if not (azop_type_spec := args[0].typeSpec()).acceptsAll(): # type check of azop was successful
			res._azop_res = azop_type_spec._azop_res # whether known or not
			if azop_type_spec._azop_args is not None: # known list of expected arguments, so we will be checking
				for k, e in edict.items():
					try:
						ckExprType(azop_type_spec._azop_args[k][1], e)
					except KeyError:
						raise AZeeError('Unknown argument "%s"' % k, src_pos)
				res._azop_args = {idict.get(k, k): tspec for k, tspec in azop_type_spec._azop_args.items() if k not in edict}
		return res, args[0].refDeps().union(*(e.refDeps() for e in edict.values()))

	elif op == 'apply': # variable return type
		if not args:
			raise AZeeError('Expected at least one argument.', src_pos)
		ckExprType(TypeSpec(Type.AZOP), args[0])
		azop_type_spec = args[0].typeSpec()
		labels, idict, edict = labelled_arg_list(args[1:])
		ref_deps = args[0].refDeps().union(*(e.refDeps() for e in edict.values()))
		if azop_type_spec.acceptsAll(): # type check of azop was unsuccessful
			return TypeSpec(None), ref_deps
		else:
			if azop_type_spec._azop_args is not None: # known list of arguments
				for param in labels:
					if param not in azop_type_spec._azop_args:
						AZeeError('Superfluous argument "%s"' % param, args[0].srcPos()).stderrWarning(AZeeError.LABELLED_ARG)
				for dep, (optflag, typspec) in azop_type_spec._azop_args.items(): # scanning arguments defined by the azop
					try:
						ckExprType(typspec, edict[dep])
					except KeyError:
						if not optflag:
							raise AZeeError('Missing argument "%s"' % dep, src_pos)
			return azop_type_spec._azop_res, ref_deps

	elif op == 'let': # variable return type
		if not args:
			raise AZeeError('Expected at least one argument.', src_pos)
		labels, idict, edict = labelled_arg_list(args[:-1])
		ref_deps = args[-1].refDeps()
		for lbl in reversed(labels):
			ref_deps.discard(lbl)
			ref_deps.update(edict[lbl].refDeps())
		return args[-1].typeSpec(), ref_deps

	elif op.startswith('?') or op.startswith('*'): # template operator
		return TypeSpec(None), set()

	else: # unknown operator
		raise AZeeError('Unknown operator: %s' % op, src_pos)







def labelled_arg_list(lst, allowNoExpr=False):
	res_args = [] # str list of labels in order (keys in following dicts are in this list)
	res_info = {} # str label -> str colon-separated suffix
	res_expr = {} # str label -> AZexpr on subsequent line
	i = 0
	while i < len(lst):
		colon_split = (lst[i] >> Type.STRING).split(':', maxsplit=1)
		label = colon_split[0].rstrip()
		if label in res_args:
			raise AZeeError('Duplicate argument label "%s"' % label, src=lst[i].srcPos())
		res_args.append(label)
		if len(colon_split) > 1:
			res_info[label] = colon_split[1].strip()
		if allowNoExpr: # a subsequent STRING will be interpreted as next arg
			if i < len(lst) - 1 and not lst[i + 1].isValue(t=Type.STRING):
				expr = lst[i + 1]
				i += 2
			else:
				expr = None
				i += 1
		else: # must read two lines
			try:
				expr = lst[i + 1]
				i += 2
			except IndexError:
				raise AZeeError('Missing expression for named argument "%s"' % label, src=lst[i].srcPos())
		if expr:
			if expr.isValue(t=Type.STRING):
				AZeeError('Associating a native STRING value with named arg "%s"' % label, src=lst[i-1].srcPos()).stderrWarning(AZeeError.LABELLED_ARG)
			res_expr[label] = expr
	return res_args, res_info, res_expr


def read_sync_block(specs, bind, posture): # CAUTION consumes list "specs"
	"""
	Specs format: block name, (sync rule with '=')*, block score
	"""
	if not specs:
		raise AZeeError('Block name expected under "sync"')
	block_name_arg = specs.pop(0)
	block_name = block_name_arg >> Type.STRING
	block_sync = {}
	while specs and specs[0].isValue(t=Type.STRING) and '=' in specs[0] >> Type.STRING:
		rule, arg = (specs.pop(0) >> Type.STRING).split('=', maxsplit=1)
		block_sync[rule] = arg
	if not specs:
		raise AZeeError('Missing contents for block "%s"' % block_name, src=block_name_arg.srcPos())
	block_contents_arg = specs.pop(0)
	block_contents = block_contents_arg.eval(bind, posture) # should be either SCORE or POSTURE
	if block_contents.isType(Type.POSTURE):
		if not ('start' in block_sync and 'end' in block_sync or 'duration' in block_sync):
			raise AZeeError('Missing duration or start/end for posture block.', src=block_name_arg.srcPos())
		if 'stretch' in block_sync:
			raise AZeeError('Posture blocks cannot stretch.', src=block_name_arg.srcPos())
		score = CstrScore(1, block_contents >> Type.POSTURE)
		score.setSourceAZexpr(block_contents_arg.deepderef(bind))
		return block_name, block_sync, score
	elif block_contents.isType(Type.SCORE):
		return block_name, block_sync, block_contents >> Type.SCORE
	raise AZeeError('Expected POSTURE or SCORE; got %s.' % Type.toStr(block_contents.type), src=block_contents_arg.srcPos())


def body_site_dist(s1, s2, posture):
	if posture is None:
		p1 = ask_for_vec3('SITE %s' % s1)
		p2 = ask_for_vec3('SITE %s' % s2)
	else:
		p1 = posture.sitePos(s1)
		p2 = posture.sitePos(s2)
	return (p2 - p1).length()


def id_side_num_str(table, args, bind, posture):
	try:
		res = args[0] >> Type.STRING
	except IndexError:
		raise AZeeError('Missing identifier.')
	for arg in args[1:]:
		val = arg.eval(bind, posture)
		if val.isType(Type.SIDE) or val.isType(Type.NUM):
			res += '_%s' % val
		else:
			raise AZeeError('Only SIDE and NUM arguments allowed.')
	if res not in table:
		raise AZeeError('Invalid bone/morph/site specification.')
	return res

def ask_for_vec3(prompt):
	print('? %s' % prompt)
	str_values = input().split(maxsplit=3)  # 3 to get rid of \r junk and give prio to arg count error msg
	if len(str_values) == 3:
		try:
			return vec3(*(float(v) for v in str_values))
		except ValueError:
			raise AZeeError('Could not convert coordinates: %s' % str_values)
	else:
		raise AZeeError('3 whitespace-separated NUM values required (X, Y, Z)')


def single_cstr_posture(cstr, args):
	"""
	build a POSTURE from a unique minimal constraint spec
	all args should be dereferenced
	"""
	return AZval(Type.POSTURE, [(cstr, args)])


external_avatar_cache = {}





###############################################
#########    Evaluation functions     #########
#########   (sorted by return type)   #########
###############################################


def op_let(args, bind, posture):
	labels, idict, edict = labelled_arg_list(args[:-1])
	new_bindings = bind.copy()
	for arg, e in edict.items():
		new_bindings[arg] = e.deepderef(new_bindings) # do not evaluate here or absolute coord's will show up in XML score specs
	return args[-1].eval(new_bindings, posture)

def op_apply(args, bind, posture):
	arg_types, arg_defaults, azop_body = args[0].eval(bind, posture) >> Type.AZOP
	call_labels, idict, edict = labelled_arg_list(args[1:])
	for dep in arg_types: # scanning arguments defined by the azop
		if dep not in call_labels:
			try:
				edict[dep] = arg_defaults[dep].deepderef(bind)
			except KeyError:
				raise AZeeError('Missing argument "%s"' % dep)
	return azop_body.deepderef(edict).eval(bind, posture)



# ----------- Return type is AZOP -------------

def op_azop(args, bind, posture):
	labels, idict, edict = labelled_arg_list(args[:-1], allowNoExpr=True)
	arg_types = {arg: TypeSpec.fromStr(idict[arg]) if arg in idict else TypeSpec(None) for arg in labels}
	arg_defaults = {arg: d.deepderef(bind) for arg, d in edict.items()}
	deref_bindings = {k: v for k, v in bind.items() if k not in labels}
	return AZval(Type.AZOP, (arg_types, arg_defaults, args[-1].deepderef(deref_bindings)))

def op_partialapply(args, bind, posture):
	azop_arg_types, azop_arg_defaults, azop_body = args[0].eval(bind, posture) >> Type.AZOP
	applied_labels, idict, edict = labelled_arg_list(args[1:], allowNoExpr=True)
	if bad := next((arg for arg in applied_labels if arg not in azop_arg_types), None):
		raise AZeeError('Unknown argument "%s"' % bad)
	res_arg_types = {idict.get(arg, arg): t for arg, t in azop_arg_types.items() if arg not in edict}
	res_arg_defaults = {idict.get(arg, arg): e for arg, e in azop_arg_defaults.items() if arg not in edict}
	deref_dict = edict.copy()
	deref_dict.update({oldname: expr.AZexpr.mkRef(newname) for oldname, newname in idict.items()})#TODO src pos for refs? FIXME what if azop_body contains ref's to newname already?
	return AZval(Type.AZOP, (res_arg_types, res_arg_defaults, azop_body.deepderef(deref_dict)))



# ----------- Return type is BONE -------------

def op_bone(args, bind, posture):
	return AZval(Type.BONE, id_side_num_str(bone_table, args, bind, posture))



# ----------- Return type is POSTURE -------------

def op_posture(args, bind, posture):
	cstr_list = []
	for arg in args:
		cstr_list.extend(arg.eval(bind, posture) >> Type.POSTURE)
	return AZval(Type.POSTURE, cstr_list)

def op_posturefor(args, bind, posture):
	it_arg, lst_arg, posture_arg = args
	curr_bind = bind.copy() # dict copy, in which iterator will be overridden
	res_list = []
	for val in lst_arg.eval(bind, posture) >> Type.LIST:
		curr_bind[it_arg >> Type.STRING] = val.deepderef(bind) # adds or hides on 1st iteration; replaces on following ones
		res_list.extend(posture_arg.eval(curr_bind, posture) >> Type.POSTURE)
	return AZval(Type.POSTURE, res_list)

def op_place(args, bind, posture):
	effector, target = args
	return single_cstr_posture('place', (effector.evalAZBodySite(bind, posture), target.deepderef(bind)))

def op_orient(args, bind, posture):
	a, b, d = args
	return single_cstr_posture('orient', (a >> Type.STRING, b.eval(bind, posture) >> Type.BONE, d.deepderef(bind)))

def op_look(args, bind, posture):
	return single_cstr_posture('look', args[0].deepderef(bind))

def op_lookat(args, bind, posture):
	return single_cstr_posture('lookat', args[0].deepderef(bind))

def op_morph(args, bind, posture):
	morph = id_side_num_str(morph_table, args[:-1], bind, posture)
	weight = args[-1].deepderef(bind)
	return single_cstr_posture('morph', (morph, weight))

def op_trill(args, bind, posture):
	axis, bone = args
	return single_cstr_posture('trill', (axis >> Type.STRING, bone.eval(bind, posture) >> Type.BONE))

def op_transpath(args, bind, posture):
	followed_bsite = args[0].evalAZBodySite(bind, posture)
	path_dynamics = args[1] >> Type.STRING # TODO use dynamics? at the moment old path type string is ignored in this place
	if len(args) == 2: # straight path
		return single_cstr_posture('transpath', (followed_bsite, 'straight', []))
	else: # bent path
		return single_cstr_posture('transpath', (followed_bsite, 'arc-vect', [args[2].deepderef(bind)]))



# ----------- Return type is SCORE -------------

def op_emptyscore(args, bind, posture):
	"""
	args: none
	returns: an empty score (single key frame at time 0; no duration)
	"""
	return AZval(Type.SCORE, CstrScore(1, []))

def op_posturescore(args, bind, posture):
	dur_arg, posture_arg = args
	return AZval(Type.SCORE, CstrScore(dur_arg.eval(bind, posture) >> Type.NUM, posture_arg.eval(bind, posture) >> Type.POSTURE))

def op_holdscore(args, bind, posture):
	"""
	args: none
	returns: an empty score whose meaning is to hold latest contraints of block whose end it is tacked on
	"""
	return AZval(Type.SCORE, HoldScore())

def op_sync(args, bind, posture):
	specs = args[:]
	block_name, block_sync, block_contents = read_sync_block(specs, bind, posture) # consumes "specs"
	res = SyncedScore(block_name, block_contents, syncRules=block_sync)
	while specs:
		block_name, block_sync, block_contents = read_sync_block(specs, bind, posture) # consumes "specs"
		res.addSyncedBlock(block_name, block_sync, block_contents)
	return AZval(Type.SCORE, res)

def op_scoreseq(args, bind, posture):
	"""
	args: separation time, AZlist of scores to sync in sequence
	"""
	separation_time_arg, score_list_arg = args
	score_list = [score.eval(bind, posture) >> Type.SCORE for score in score_list_arg.eval(bind, posture) >> Type.LIST]
	if len(score_list) == 0:
		return op_emptyscore([], bind, posture)
	else:
		res = SyncedScore('SCORE_0', score_list[0]) # partial SyncedScore, further built below
		for i, score in enumerate(score_list[1:], 1):
			res.addSyncedBlock('SCORE_%i' % i, {'start': 'SCORE_%i:end:%f' % (i - 1, separation_time_arg >> Type.NUM)}, score)
	return AZval(Type.SCORE, res)

def op_repeatscore(args, bind, posture):
	n, sep, score = args # TODO check n is integer > 0
	res = SyncedScore('SCORE_0', score.eval(bind, posture) >> Type.SCORE)
	for i in range(1, int(n.eval(bind, posture) >> Type.NUM)):
		res.addSyncedBlock('SCORE_%i' % i, {'start': 'SCORE_%i:end:%f' % (i - 1, sep >> Type.NUM)}, score.eval(bind, posture) >> Type.SCORE)
	return AZval(Type.SCORE, res)

def op_dynamicpoint(args, bind, posture):
	pt, num, path, dyn, score = args
	point_name = pt >> Type.STRING
	num_name = num >> Type.STRING
	deref_dict = {k: v for k, v in bind.items() if k != point_name and k != num_name}
	res = score.eval(deref_dict, posture) >> Type.SCORE
	res.addDynamicPoint(point_name, num_name, path.deepderef(deref_dict), dyn >> Type.STRING)
	return AZval(Type.SCORE, res)

def op_movesite(args, bind, posture):
	dur, site, dyn, path = args
	implicit_dynpt = '_pt'
	implicit_dynnum = '_t'
	implicit_place = 'place', (site.evalAZBodySite(bind, posture), expr.AZexpr.mkRef(implicit_dynpt))
	res = CstrScore(dur.eval(bind, posture) >> Type.NUM, [implicit_place])
	res.addDynamicPoint(implicit_dynpt, implicit_dynnum, path.deepderef(bind), dyn >> Type.STRING)
	return AZval(Type.SCORE, res)



# ----------- Return type is SIDE -------------

def op_oppside(args, bind, posture):
	return AZval(Type.SIDE, not args[0].eval(bind, posture) >> Type.SIDE)

def op_l(args, bind, posture):
	return AZval(Type.SIDE, False)

def op_r(args, bind, posture):
	return AZval(Type.SIDE, True)

def op_s(args, bind, posture):
	return AZval(Type.SIDE, strong_side)

def op_w(args, bind, posture):
	return AZval(Type.SIDE, not strong_side)



# ----------- Return type is VECT -------------

def op_dir(args, bind, posture):
	b = args[0].eval(bind, posture) >> Type.BONE
	if posture is None:
		return AZval(Type.VECT, ask_for_vec3('DIR %s' % b))
	else:
		return AZval(Type.VECT, posture.dirVect(bone=b))

def op_nrm(args, bind, posture):
	b = args[0].eval(bind, posture) >> Type.BONE
	if posture is None:
		return AZval(Type.VECT, ask_for_vec3('NRM %s' % b))
	else:
		return AZval(Type.VECT, posture.nrmVect(bone=b))

def op_RGT(args, bind, posture):
	try:
		v = external_avatar_cache[RGT_VECT_CACHE_KEY]
	except KeyError:
		vec3_to_addr = (op_addr([], bind, posture) >> Type.POINT) - (op_Mssp([], bind, posture) >> Type.POINT)
		v = external_avatar_cache[RGT_VECT_CACHE_KEY] = vec3_to_addr.cross(op_up([], bind, posture) >> Type.VECT)
	return AZval(Type.VECT, v)

def op_LFT(args, bind, posture):
	return AZval(Type.VECT, -(op_RGT([], bind, posture) >> Type.VECT))

def op_LAT(args, bind, posture):
	return op_RGT([], bind, posture) if strong_side else op_LFT([], bind, posture)

def op_CONTRA(args, bind, posture):
	return AZval(Type.VECT, -(op_LAT([], bind, posture) >> Type.VECT))

def op_FWD(args, bind, posture):
	try:
		v = external_avatar_cache[FWD_VECT_CACHE_KEY]
	except KeyError:
		v = external_avatar_cache[FWD_VECT_CACHE_KEY] = (op_up([], bind, posture) >> Type.VECT).cross(op_RGT([], bind, posture) >> Type.VECT)
	return AZval(Type.VECT, v)

def op_BWD(args, bind, posture):
	return AZval(Type.VECT, -(op_FWD([], bind, posture) >> Type.VECT))

def op_up(args, bind, posture):
	try:
		v = external_avatar_cache[UP_VECT_CACHE_KEY]
	except KeyError:
		v = ask_for_vec3(UP_VECT_CACHE_KEY) if posture is None else posture.upVect()
		external_avatar_cache[UP_VECT_CACHE_KEY] = v
	return AZval(Type.VECT, v)

def op_dwn(args, bind, posture):
	return AZval(Type.VECT, -(op_up([], bind, posture) >> Type.VECT))

def op_rgt(args, bind, posture):
	return AZval(Type.VECT, (op_nrm([expr.AZexpr.fromValue(Type.BONE, 'back_2')], bind, posture) >> Type.VECT).cross(op_up([], bind, posture) >> Type.VECT))

def op_lft(args, bind, posture):
	return AZval(Type.VECT, -(op_rgt([], bind, posture) >> Type.VECT))

def op_fwd(args, bind, posture):
	return AZval(Type.VECT, (op_up([], bind, posture) >> Type.VECT).cross(op_rgt([], bind, posture) >> Type.VECT))

def op_bwd(args, bind, posture):
	return AZval(Type.VECT, -(op_fwd([], bind, posture) >> Type.VECT))

def op_lat(args, bind, posture):
	return op_rgt([], bind, posture) if strong_side else op_lft([], bind, posture)

def op_contra(args, bind, posture):
	return AZval(Type.VECT, -(op_lat([], bind, posture) >> Type.VECT))

def op_sidevect(args, bind, posture):
	s, v = args
	vect = v.eval(bind, posture) >> Type.VECT
	return AZval(Type.VECT, (vect if s.eval(bind, posture) >> Type.SIDE else -vect))

def op_rhvect(args, bind, posture):
	v = args[0].eval(bind, posture) >> Type.VECT
	return AZval(Type.VECT, v) if strong_side else AZval(Type.VECT, -v)

def op_oppvect(args, bind, posture):
	return AZval(Type.VECT, -(args[0].eval(bind, posture) >> Type.VECT))

def op_addvect(args, bind, posture):
	va, vb = args
	return AZval(Type.VECT, (va.eval(bind, posture) >> Type.VECT) + (vb.eval(bind, posture) >> Type.VECT))

def op_scalevect(args, bind, posture):
	k, v = args
	return AZval(Type.VECT, (v.eval(bind, posture) >> Type.VECT) * (k.eval(bind, posture) >> Type.NUM))

def op_crossprod(args, bind, posture):
	va, vb = args
	return AZval(Type.VECT, (va.eval(bind, posture) >> Type.VECT).cross(vb.eval(bind, posture) >> Type.VECT))

def op_along(args, bind, posture):
	v, x = args
	vdir = v.eval(bind, posture) >> Type.VECT
	return AZval(Type.VECT, vdir * ((x.eval(bind, posture) >> Type.NUM) / vdir.length()))

def op_vect(args, bind, posture):
	pa, pb = args
	return AZval(Type.VECT, (pb.eval(bind, posture) >> Type.POINT) - (pa.eval(bind, posture) >> Type.POINT))

def op_tangentonpath(args, bind, posture):
	p, n = args
	path = p.eval(bind, posture) >> Type.PATH
	num = n.eval(bind, posture) >> Type.NUM
	return AZval(Type.VECT, (path.pointOnPath(num + 1e-6) - path.pointOnPath(num - 1e-6)).normalized())



# ----------- Return type is POINT -------------

def op_Mssp(args, bind, posture):
	try:
		mssp = external_avatar_cache[MSSP_POINT_CACHE_KEY]
	except KeyError:
		mssp = ask_for_vec3(MSSP_POINT_CACHE_KEY) if posture is None else posture.MsspPos()
		external_avatar_cache[MSSP_POINT_CACHE_KEY] = mssp
	return AZval(Type.POINT, mssp)

def op_addr(args, bind, posture):
	try:
		addr = external_avatar_cache[ADDR_POINT_CACHE_KEY]
	except KeyError:
		addr = ask_for_vec3(ADDR_POINT_CACHE_KEY) if posture is None else posture.addresseePos()
		external_avatar_cache[ADDR_POINT_CACHE_KEY] = addr
	return AZval(Type.POINT, addr)

def op_site(args, bind, posture):
	"""
	Returns the POINT position of the body site in the skeleton spec.
	"""
	site = id_side_num_str(site_table, args, bind, posture)
	if posture is None:
		return AZval(Type.POINT, ask_for_vec3('SITE %s' % site))
	else:
		return AZval(Type.POINT, posture.sitePos(site))

def op_origin(args, bind, posture):
	b = args[0].eval(bind, posture) >> Type.BONE
	if posture is None:
		return AZval(Type.POINT, ask_for_vec3('ORIGIN %s' % b))
	else:
		return AZval(Type.POINT, posture.jointPos(b))

def op_translate(args, bind, posture):
	pt, v, dist = args
	vect = v.eval(bind, posture) >> Type.VECT
	vect.normalize()
	vect *= dist.eval(bind, posture) >> Type.NUM
	return AZval(Type.POINT, (pt.eval(bind, posture) >> Type.POINT) + vect)

def op_vtranslate(args, bind, posture):
	pt = args[0].eval(bind, posture) >> Type.POINT
	for v in args[1:]:
		pt += v.eval(bind, posture) >> Type.VECT
	return AZval(Type.POINT, pt)

def op_midpoint(args, bind, posture):
	p1, p2 = args
	p = p1.eval(bind, posture) >> Type.POINT
	v = (p2.eval(bind, posture) >> Type.POINT) - p
	return AZval(Type.POINT, p + v * .5)

def op_pointonpath(args, bind, posture):
	p, n = args
	return AZval(Type.POINT, (p.eval(bind, posture) >> Type.PATH).pointOnPath(n.eval(bind, posture) >> Type.NUM))

def op_pointsym(args, bind, posture):
	centre, point = args
	c = centre.eval(bind, posture) >> Type.POINT
	p = point.eval(bind, posture) >> Type.POINT
	return AZval(Type.POINT, c * 2 - p)

def op_linesym(args, bind, posture):
	pline, dirline, point = args
	c = pline.eval(bind, posture) >> Type.POINT
	d = dirline.eval(bind, posture) >> Type.VECT
	p = point.eval(bind, posture) >> Type.POINT
	# H := orthogonal projection of P on axis; vect(CH) = (xM+yM+zM-xC-yC-zC) * nrm(d)
	# k is the float number such that vect(PH) = k d
	k = d * (p - c) / (d * d)
	orth_proj = c + d * k
	return AZval(Type.POINT, orth_proj * 2 - p)

def op_planesym(args, bind, posture):
	pplane, nrmplane, point = args
	c = pplane.eval(bind, posture) >> Type.POINT
	n = nrmplane.eval(bind, posture) >> Type.VECT
	p = point.eval(bind, posture) >> Type.POINT
	# H := orthogonal projection of P on plane
	# k is such that vect(PH) = k v
	k = n * (c - p) / (n * n)
	orth_proj = p + n * k
	return AZval(Type.POINT, orth_proj * 2 - p)



# ----------- Return type is PATH -------------

def op_straightfrom(args, bind, posture):
	p, v = args
	start = p.eval(bind, posture) >> Type.POINT
	return AZval(Type.PATH, StraightPath(start, start + (v.eval(bind, posture) >> Type.VECT)))

def op_straightto(args, bind, posture):
	v, p = args
	end = p.eval(bind, posture) >> Type.POINT
	return AZval(Type.PATH, StraightPath(end - (v.eval(bind, posture) >> Type.VECT), end))

def op_straightthrough(args, bind, posture):
	v, p = args
	through = p.eval(bind, posture) >> Type.POINT
	vect = v.eval(bind, posture) >> Type.VECT
	return AZval(Type.PATH, StraightPath(through - .5 * vect, through + .5 * vect))

def op_straightbetween(args, bind, posture):
	p1, p2 = args
	return AZval(Type.PATH, StraightPath(p1.eval(bind, posture) >> Type.POINT, p2.eval(bind, posture) >> Type.POINT))

def op_arcfrom(args, bind, posture):
	p, c, n, r = args
	start = p.eval(bind, posture) >> Type.POINT
	centre = c.eval(bind, posture) >> Type.POINT
	nrm_vect = n.eval(bind, posture) >> Type.VECT
	rev_count = r.eval(bind, posture) >> Type.NUM
	return AZval(Type.PATH, ArcPath(centre, start, nrm_vect, rev_count))

def op_arcthrough(args, bind, posture):
	v, c, n, r = args
	thru_point_vect = v.eval(bind, posture) >> Type.VECT
	centre = c.eval(bind, posture) >> Type.POINT
	nrm_vect = n.eval(bind, posture) >> Type.VECT
	rev_count = r.eval(bind, posture) >> Type.NUM
	return AZval(Type.PATH, ArcPath(centre, thru_point_vect.rot(quat.rotation(nrm_vect, -pi * rev_count)) + centre, nrm_vect, rev_count))

def op_arcto(args, bind, posture):
	p, c, n, r = args
	end_pt = p.eval(bind, posture) >> Type.POINT
	centre = c.eval(bind, posture) >> Type.POINT
	nrm_vect = n.eval(bind, posture) >> Type.VECT
	rev_count = r.eval(bind, posture) >> Type.NUM
	return AZval(Type.PATH, ArcPath(centre, (end_pt - centre).rot(quat.rotation(nrm_vect, -2 * pi * rev_count)) + centre, nrm_vect, rev_count))

def op_wavyfrom(args, bind, posture):
	p, v, n, a, w = args
	start = p.eval(bind, posture) >> Type.POINT
	vect_to_end = v.eval(bind, posture) >> Type.VECT
	nrm_vect = n.eval(bind, posture) >> Type.VECT
	amplitude = a.eval(bind, posture) >> Type.NUM
	wave_count = w.eval(bind, posture) >> Type.NUM
	return AZval(Type.PATH, SinePath(start, vect_to_end, nrm_vect, amplitude, wave_count, 0))

def op_curve(args, bind, posture):
	return AZval(Type.PATH, CatmullRomPath([p.eval(bind, posture) >> Type.POINT for p in args[0].eval(bind, posture) >> Type.LIST]))

def op_vcurve(args, bind, posture):
	curve_points = [args[0].eval(bind, posture) >> Type.POINT]
	for v in args[1:]:
		curve_points.append(curve_points[-1] + (v.eval(bind, posture) >> Type.VECT))
	return AZval(Type.PATH, CatmullRomPath(curve_points))



# ----------- Return type is NUM -------------

def op_opp(args, bind, posture):
	return AZval(Type.NUM, -(args[0].eval(bind, posture) >> Type.NUM))

def op_plus(args, bind, posture):
	a, b = args
	return AZval(Type.NUM, (a.eval(bind, posture) >> Type.NUM) + (b.eval(bind, posture) >> Type.NUM))

def op_mult(args, bind, posture):
	a, b = args
	return AZval(Type.NUM, (a.eval(bind, posture) >> Type.NUM) * (b.eval(bind, posture) >> Type.NUM))

def op_div(args, bind, posture):
	a, b = args
	return AZval(Type.NUM, (a.eval(bind, posture) >> Type.NUM) / (b.eval(bind, posture) >> Type.NUM))

def op_dist(args, bind, posture):
	pa, pb = args
	vect_ab = (pb.eval(bind, posture) >> Type.POINT) - (pa.eval(bind, posture) >> Type.POINT)
	return AZval(Type.NUM, vect_ab.length())

def op_tiny(args, bind, posture):
	try:
		num = external_avatar_cache[TINY_DIST_CACHE_KEY]
	except KeyError:
		num = external_avatar_cache[TINY_DIST_CACHE_KEY] = body_site_dist('L_KN_R_3', 'L_TIP_R', posture)
	return AZval(Type.NUM, num)

def op_small(args, bind, posture):
	try:
		num = external_avatar_cache[SMALL_DIST_CACHE_KEY]
	except KeyError:
		num = external_avatar_cache[SMALL_DIST_CACHE_KEY] = .5 * (op_medium([], bind, posture) >> Type.NUM)
	return AZval(Type.NUM, num)

def op_medium(args, bind, posture):
	try:
		num = external_avatar_cache[MEDIUM_DIST_CACHE_KEY]
	except KeyError:
		num = external_avatar_cache[MEDIUM_DIST_CACHE_KEY] = body_site_dist('EL_R', 'WR_R', posture)
	return AZval(Type.NUM, num)

def op_large(args, bind, posture):
	try:
		num = external_avatar_cache[LARGE_DIST_CACHE_KEY]
	except KeyError:
		num = external_avatar_cache[LARGE_DIST_CACHE_KEY] = 1.75 * (op_medium([], bind, posture) >> Type.NUM)
	return AZval(Type.NUM, num)

def op_huge(args, bind, posture):
	try:
		num = external_avatar_cache[HUGE_DIST_CACHE_KEY]
	except KeyError:
		num = external_avatar_cache[HUGE_DIST_CACHE_KEY] = 2.25 * (op_medium([], bind, posture) >> Type.NUM)
	return AZval(Type.NUM, num)



# ----------- Return type is LIST -------------

## CAUTION
## These are all lists OF DEREF'ED EXPRESSIONS (not fully evaluated elements).
## Breaking this fails to keep references in source expr printing of nested scores.

def op_list(args, bind, posture):
	"""
	args: AZexpr list ('args' is the list, this is different to 1 AZlist arg)
	List does not necessarily return every argument evaluated, only the top-level
	evaluation evaluates each arg.
	"""
	return AZval(Type.LIST, [e.deepderef(bind) for e in args])

def op_listconcat(args, bind, posture):
	return AZval(Type.LIST, [e.deepderef(bind) for lst_expr in args for e in lst_expr.eval(bind, posture) >> Type.LIST])

def op_listfor(args, bind, posture):
	"""
	args: str iterator, expr list, AZexpr
	creates an AZlist of expressions from an iteration over values
	"""
	it_arg, lst_arg, res_arg = args
	curr_bind = bind.copy() # dict copy, in which iterator will be overridden
	res_list = []
	for val in lst_arg.eval(bind, posture) >> Type.LIST:
		curr_bind[it_arg >> Type.STRING] = val.deepderef(bind) # adds or hides on 1st iteration; replaces on following ones
		res_list.append(res_arg.deepderef(curr_bind))
	return AZval(Type.LIST, res_list)

def op_repeatlist(args, bind, posture):
	"""
	args: num N, expr E
	Creates a list of N times the value E
	"""
	n, e = args # TODO check n is integer > 0
	return AZval(Type.LIST, int(n.eval(bind, posture) >> Type.NUM) * [e.deepderef(bind)])
