# -*- coding: utf-8 -*-

from azee.util.debug import *
from azee.util.xml import XML_output

from azee.values import AZeeError


default_XML_includes_flattened_scores = False


def cstr_xml(cstr):
    cstr_type, args = cstr
    if cstr_type == 'place':
        bs, te = args
        return '<place effector="%s">%s</place>' % (bs, te.pipedStr())
    elif cstr_type == 'orient':
        ax, bn, dv = args
        return '<orient bone="%s" axis="%s">%s</orient>' % (bn, ax, dv.pipedStr())
    elif cstr_type == 'look' or cstr_type == 'lookat':
        return '<%s>%s</%s>' % (cstr_type, args.pipedStr(), cstr_type)
    elif cstr_type == 'morph':
        morph, weight = args
        return '<morph id="%s">%s</morph>' % (morph, weight.pipedStr())
    elif cstr_type == 'trill':
        ax, bn = args
        return '<trill bone="%s" axis="%s"/>' % (bn, ax)
    elif cstr_type == 'transpath':
        bs, ptype, pargs = args
        return '<transpath effector="%s" pathtype="%s"> %s </transpath>' % (bs, ptype, ' '.join('<param>%s</param>' % param.pipedStr() for param in pargs))
    else:
        return '<!-- "%s" constraint -->' % cstr_type



## ABSTRACT SCORE CLASSES

# Base class for ALL scores
class Score:
	def __init__(self):
		self.source = None # optional info: the AZexpr that led to building this score
		self.dynamic_point = None
	
	def setSourceAZexpr(self, expr):
		self.source = expr
	
	def sourceExpr(self):
		return self.source

	def blockCount(self): # TO OVERRIDE
		raise NotImplementedError('Abstract method not overridden: blockCount')
	
	def flattened(self): # TO OVERRIDE, should return a fresh NonAnimatedKeyFrameScore
		raise NotImplementedError('Abstract method not overridden: flattened')
	
	def xml(self): # TO OVERRIDE
		raise NotImplementedError('Abstract method not overridden: xml')
	
	def addDynamicPoint(self, point_name, num_name, path_expr, dynamics_type):
		self.dynamic_point = point_name, num_name, path_expr, dynamics_type
	
	def includeXmlDynamicPoint(self, io):
		if self.dynamic_point is not None:
			p, x, path_expr, dynamics = self.dynamic_point
			io.open_tag('DynSpec')
			io.open_tag('PointPath', attr=[('pointName', p), ('trackingNum', x)])
			io.indentln(path_expr.pipedStr())
			io.close_tag() # </PointPath>
			io.line_tag('Dynamics', attr=[('style', dynamics)])
			io.close_tag() # </DynSpec>
	
	def includeXmlSource(self, io):
		if e := self.sourceExpr():
			io.open_tag('AZexpr')
			io.indentln(e.deepderef({}, derefLet=True).pipedStr())
			io.close_tag() # </AZexpr>



####################################################################



class CstrScore(Score):
	def __init__(self, duration, constraints):
		Score.__init__(self)
		self.duration = max(1e-12, duration)
		self.constraints = constraints

	def blockCount(self):
		return 1
	
	def flattened(self):
		res = NonAnimatedKeyFrameScore()
		start = res.addKeyFrame(0)
		end = res.addKeyFrame(self.duration)
		res.addCstr(start, end, self.constraints)
		return res
	
	def xml(self):
		io = XML_output('  ')
		io.open_tag('CstrScore')
		self.includeXmlSource(io)
		self.includeXmlDynamicPoint(io)
		io.line_tag('duration', [('hint', str(self.duration))])
		io.open_tag('CstrSpec')
		for cstr in self.constraints:
			io.indentln(cstr_xml(cstr))
		io.close_tag() # /CstrSpec
		io.close_tag() # /CstrScore
		return io.getvalue()




class HoldScore(Score):
	def __init__(self):
		Score.__init__(self)
		self.constraints = []
	
	def addHeldConstraints(self, constraints):
		self.constraints.extend(constraints)

	def blockCount(self):
		return 1
	
	def flattened(self):
		res = NonAnimatedKeyFrameScore()
		start = res.addKeyFrame(0)
		end = res.addKeyFrame(1) # dummy timing
		res.addCstr(start, end, self.constraints)
		return res
	
	def xml(self):
		io = XML_output('  ')
		io.open_tag('HoldScore')
		self.includeXmlSource(io)
		self.includeXmlDynamicPoint(io)
		io.open_tag('CstrSpec')
		for cstr in self.constraints:
			io.indentln(cstr_xml(cstr))
		io.close_tag() # /CstrSpec
		io.close_tag() # /CstrScore
		return io.getvalue()




class SyncedScore(Score):
	def __init__(self, base_block, base_contents, syncRules=None):
		Score.__init__(self)
		self.sync_order = [base_block] # names of blocks synced after base block
		self.sync_rules = {base_block: syncRules if syncRules else {}} # str block name -> (sync key -> value)
		self.block_contents = {base_block: base_contents} # str -> score
		# Flattened stuff, building up as new blocks are later added
		self._flattenend = base_contents.flattened()
		if 'stretch' in self.sync_rules:
			try:
				self._flattenend.stretch(float(self.sync_rules['stretch']))
			except ValueError:
				raise AZeeError('Bad "stretch" value for block %s' % base_block)
		self._block_keyframes = {base_block: self._flattenend.orderedKeyFrames()}
		#trace('Base result score with duration %g and %d keyframes' % \
		#	(self.flat_score_eval.duration(), self.flat_score_eval.keyFrameCount()))

	def syncedBlocks(self):
		return self.sync_order[:]

	def blockCount(self):
		return sum(block.blockCount() for block in self.block_contents.values())

	def flattened(self):
		return self._flattenend
	
	def addSyncedBlock(self, block_name, sync_rules, contents):
		if block_name in self.sync_order:
			raise AZeeError('Duplicate block name: %s' % block_name)
		sync = sync_rules.copy()
		
		# Add new block to score
		self.sync_order.append(block_name)
		self.sync_rules[block_name] = sync
		self.block_contents[block_name] = contents
		
		# Update flattened score
		flattened_contents = contents.flattened() # returns a fresh new object
		kf1 = kf2 = None
		if isinstance(contents, HoldScore):
			if 'after' in sync:
				kf1 = self._getFlatKeyframe(sync['after'] + ':end')
			elif 'before' in sync:
				kf2 = self._getFlatKeyframe(sync['before'] + ':start')
		if 'start' in sync:
			kf1 = self._getFlatKeyframe(sync['start'])
		if 'end' in sync:
			kf2 = self._getFlatKeyframe(sync['end'])
		if kf1 is None and kf2 is None:
			raise AZeeError('Start or end needed to sync block %s' % block_name)

		try:
			dur = float(sync['duration']) # CHECKME ValueError caught? vs. "stretch" rule below
			if 'stretch' in sync:
				raise AZeeError('Cannot stretch block %s; duration given' % block_name)
		except KeyError: # no duration rule given
			dur = None

		if kf1 is not None and kf2 is not None: # both start/end explicitly given
			if dur is not None or 'stretch' in sync:
				raise AZeeError('Cannot force duration or stretch block %s; both start and end given' % block_name)
		elif isinstance(contents, HoldScore): # has no natural internal duration
			if dur is None:
				raise AZeeError('Missing duration or start/end for hold block %s' % block_name)
		elif dur is None: # has no natural internal duration
			dur = flattened_contents.duration()
			try:
				dur *= float(sync['stretch'])
			except KeyError:
				pass
			except ValueError:
				raise AZeeError('Bad "stretch" value for block %s' % block_name)

		if kf1 is None:
			kf1 = self._flattenend.addKeyFrame(self._flattenend.t(kf2) - dur)
		elif kf2 is None:
			kf2 = self._flattenend.addKeyFrame(self._flattenend.t(kf1) + dur)
		# add to the flat score and remember final/initial constraints so that we know what to hold if the block is later held
		self._block_keyframes[block_name] = self._flattenend.fitBetween(kf1, kf2, flattened_contents)
	
	def _getFlatKeyframe(self, strspec):
		tokens = strspec.split(':', maxsplit=2)
		if len(tokens) < 2:
			raise AZeeError('Bad time spec "%s" (expected BLOCK:boundary[:offset])' % strspec)
		ti, boundary = tokens[0:2]
		nthkf = {'start': 0, 'end': -1}[boundary]
		try:
			kf = self._block_keyframes[ti][nthkf]
			if len(tokens) == 3:
				off = float(tokens[2])
				if off != 0:
					kf = self._flattenend.addKeyFrame(self._flattenend.t(kf) + off)
			return kf
		except KeyError:
			raise AZeeError('Unidentified block "%s" used in sync rule' % ti)
	
	def xml(self, includeFlattened=default_XML_includes_flattened_scores):
		io = XML_output('  ')
		io.open_tag('ScoreSync')
		io.line_comment('ScoreSync contains %d blocks: %s' % (len(self.sync_order), ', '.join(self.sync_order)))
		self.includeXmlSource(io)
		self.includeXmlDynamicPoint(io)
		io.open_tag('blocks')
		for block_name in self.sync_order:
			block_contents = self.block_contents[block_name]
			io.open_tag('block', [('name', block_name)])
			io.indent_text(block_contents.xml())
			io.close_tag() # /block
		io.close_tag() # /blocks
		io.open_tag('sync_rules')
		for block_name, sync_rule in self.sync_rules.items():
			for attr, value in sync_rule.items():
				io.line_tag('sync', [('block', block_name), ('attr', attr), ('value', value)])
		io.close_tag() # /sync_rules
		if includeFlattened:
			io.indent_text(self._flattenend.xml())
		io.close_tag() # /ScoreSync
		return io.getvalue()











####################################################################



# Base class for non-nested (flat) scores
class KeyFrameScore(Score): # NOTE: remains ABSTRACT; flattened() and xml() implementations still needed
	def __init__(self):
		Score.__init__(self)
		self.kf_times = {}    # dict: kfid -> time
		self.kf_comments = {} # dict: kfid -> str


	#---------------#
	#   Accessors   #
	#---------------#

	def keyFrameCount(self):
		return len(self.kf_times)

	def t(self, kf):
		return self.kf_times[kf]

	def duration(self):
		return self.endTime() - self.startTime()

	def startTime(self):
		return min(self.kf_times.values())

	def endTime(self):
		return max(self.kf_times.values())

	def firstKeyFrame(self):
		return min(self.kf_times, key=self.kf_times.get)

	def lastKeyFrame(self):
		return max(self.kf_times, key=self.kf_times.get)

	def orderedKeyFrames(self):
		return sorted(self.kf_times, key=self.kf_times.get)

	def comment(self, kf):
		return self.kf_comments[kf]


	#-------------------#
	#   KF arithmetic   #
	#-------------------#

	def before(self, kf1, kf2, strict=False):
		return self.t(kf1) < self.t(kf2) or not strict and kf1 == kf2

	def inInterval(self, kf, interval, strict=False):
		return self.before(interval[0], kf, strict) and self.before(kf, interval[1], strict)

	def intervalOverlaps(self, ab, cd, strict):
		abocd = self.inInterval(cd[0], ab, strict=False) and not (strict and ab[1] == cd[0])
		return abocd or self.inInterval(ab[0], cd, strict=False) and not (strict and ab[0] == cd[1])


	#---------------#
	#   Modifiers   #
	#---------------#

	def addKeyFrame(self, time, comment=None):
		time = float(time)
		if time in self.kf_times.values():
			trace('WARNING: keyframe exists at time %s' % time, channel='eval-score')
		keyID = 0
		while keyID in self.kf_times:
			keyID += 1
		self.kf_times[keyID] = time
		self.kf_comments[keyID] = comment
		return keyID

	def offset(self, off):
		self.kf_times = {kf: t + off for kf, t in self.kf_times.items()}

	def stretch(self, stretch_factor):
		self.kf_times = {kf: stretch_factor * t for kf, t in self.kf_times.items()}

	def fitTimeWindow(self, start, end):
		stretch = (end - start) / self.duration()
		pad = start - stretch * self.startTime()
		self.kf_times = {kf: stretch * t + pad for kf, t in self.kf_times.items()}




####################################################################



class NonAnimatedKeyFrameScore(KeyFrameScore):
	"""
	Avatar animation specification containing AZee constraints held between keyframes.
	"""
	def __init__(self):
		KeyFrameScore.__init__(self)
		self.AZee_cstr = [] # list of ((kfid, kfid), cstr)
	
	def flattened(self): # implements abstract method by returning a copy of the flat KF score built on the fly
		result = NonAnimatedKeyFrameScore()
		result.kf_times = self.kf_times.copy()
		result.kf_comments = self.kf_comments.copy()
		result.AZee_cstr = [(interval, cstrlst[:]) for interval, cstrlst in self.AZee_cstr]
		return result

	def constraintsAtKeyFrame(self, kf):
		return [cstr for interval, cstrlst in self.AZee_cstr if self.inInterval(kf, interval) for cstr in cstrlst]


	#---------------#
	#   Modifiers   #
	#---------------#

	def addCstr(self, kf_start, kf_end, cstr):
		assert isinstance(cstr, list)
		assert self.before(kf_start, kf_end)
		self.AZee_cstr.append(((kf_start, kf_end), cstr))


	#-------------------------------#
	#   Merging with other scores   #
	#-------------------------------#

	def _transferSpecs(self, other, keytbl):
		for (i1, i2), cstrlst in other.AZee_cstr:
			self.addCstr(keytbl[i1], keytbl[i2], cstrlst[:])
		for okf, cmt in other.kf_comments.items():
			self.kf_comments[keytbl[okf]] = cmt

	def fitBetween(self, kf_start, kf_end, other):
		"""
		Other score is inserted between given key frames of this score. Time is scaled and offset
		to fit the required time window, but relative durations in 'other' are unchanged.
		Other score must not be empty; interval(kf_start, kf_end) must exist in this score.
		Returns the ordered list of renamed keyframes corresponding the original keyframe IDs in other.
		"""
		okfs = other.orderedKeyFrames()
		#trace('Fitting keyframes %s between (%s:%g) and (%s:%g)' % (okfs, kf_start, self.t(kf_start), kf_end, self.t(kf_end)))
		ostretch = (self.t(kf_end) - self.t(kf_start)) / (other.t(okfs[-1]) - other.t(okfs[0]))
		opad = self.t(kf_start) - ostretch * other.t(okfs[0])
		keytbl = {okfs[0]: kf_start, okfs[-1]: kf_end}
		#trace('%s' % keytbl)
		for okf in okfs[1:-1]:
			keytbl[okf] = self.addKeyFrame(other.t(okf) * ostretch + opad)
		self._transferSpecs(other, keytbl)
		return [keytbl[okf] for okf in okfs]


	def alignFrom(self, kf_start, other):
		"""
		Other score is inserted starting at given key frame of this score, aligning timeline keyframes
		in order of both scores. Timings are ignored in 'other' as long as there are keyframes to align
		with in this score; 'other' durations are transfered as further keyframes are needed and inserted.
		Returns the ordered list of renamed keyframes corresponding the original keyframe IDs in other.
		"""
		kfs = self.orderedKeyFrames()
		tail_timeline = kfs[kfs.index(kf_start)+1:]
		okfs = other.orderedKeyFrames()
		keytbl = {okfs[0]: kf_start}
		for okf in okfs[1:]:
			if tail_timeline:
				keytbl[okf] = tail_timeline.pop(0)
			else:
				keytbl[okf] = self.addKeyFrame(self.t(kf_start) - other.t(okfs[0]) + other.t(okf))
		self._transferSpecs(other, keytbl)
		return [keytbl[okf] for okf in okfs]


	#-------------------------#
	#   String descriptions   #
	#-------------------------#

	def xml(self):
		io = XML_output('  ')
		io.open_tag('KeyFrameScore')
		io.line_comment('Score contains %d key frames.' % self.keyFrameCount())
		io.open_tag('KeyFrames')
		for key in sorted(self.kf_times, key=self.t):
			io.line_tag('KeyFrame', [('id', key), ('time', self.t(key))])
		io.close_tag() # /KeyFrames
		io.open_tag('AnimSpec')
		for (start, end), cstrlst in sorted(self.AZee_cstr, key=(lambda x: x[0][0])):
			io.open_tag('hold', [('start', start), ('end', end)])
			for cstr in cstrlst:
				io.indentln(cstr_xml(cstr))
			io.close_tag() # /hold
		io.close_tag() # /AnimSpec
		io.close_tag() # /NonAnimatedScore
		return io.getvalue()
