
import re

from azee.util.debug import *
from azee.values import Type, TypeSpec, AZeeError
from azee.dbase import known_ref, local_db_deref
from azee.posture import dummy_posture

prod_rules_dir_name = 'production-set'

int_comp_op_regex = re.compile('(?P<op>[a-z]+)(?P<comp>=|<>|<|<=|>|>=)(?P<val>\\d+)$')



class TemplateMatch:
    def __init__(self, matched_expr, wilcards=None):
        self.matched_expr = matched_expr
        self.wildcard_matches = {} if wilcards is None else wilcards

    def matchedExpr(self):
        return self.matched_expr

    def matchedWildcard(self, wc):
        return self.wildcard_matches[wc]

    def wildcardDict(self):
        return self.wildcard_matches.copy()

    def wildcards(self):
        return self.wildcard_matches.keys()

    def extendWith(self, other):
        for wildcard, matched_expr in other.wildcard_matches.items():
            if wildcard in self.wildcard_matches:
                raise AZeeError('Multiple use of wilcard "%s" in tested template' % wildcard)
            self.wildcard_matches[wildcard] = matched_expr

    def __str__(self):
        return 'TemplateMatch-{' + ', '.join('%s: %s' % wcmap for wcmap in self.wildcard_matches.items()) + '}'



def int_comp_op_test(test_op, tested_value):
    match = int_comp_op_regex.match(test_op)
    if match:
        op_value = int(match.group('val'))
        f_comp = {
            '=': tested_value.__eq__, '<': tested_value.__lt__, '<=': tested_value.__le__,
            '<>': tested_value.__ne__, '>': tested_value.__gt__, '>=': tested_value.__ge__
        }[match.group('comp')]
        return f_comp(op_value)
    else:
        raise AZeeError('Invalid syntax for int comparison test: ' + test_op)



def prod_node(expr):
    if rule := expr.possibleInfamousLexicalFallback(): # can be None
        return rule
    if (expr.isOperator('apply') or expr.isOperator('partialapply')) and len(expr.args()) > 0 and (ref := expr.arg(0).reference()):
        if known_ref(ref):
            src_pos = local_db_deref(ref).srcPos()
            if src_pos is not None and src_pos.file and prod_rules_dir_name in src_pos.file:
                return ref
    return None

def prod_args(expr):
    if prod_node(expr):
        return [arg for arg in expr.args()[1:] if not arg.isValue(t=Type.STRING)]
    elif expr.isOperator():
        return expr.args()
    else:
        return []

def prod_count(expr):
    return int(prod_node(expr) is not None) + sum(prod_count(arg) for arg in prod_args(expr))


def prod_height(expr):
    return int(prod_node(expr) is not None) + max((prod_height(arg) for arg in prod_args(expr)), default=0)


def expr_contains(expr, test):
    return test.templateMatch(expr) or any(expr_contains(arg, test) for arg in prod_args(expr))



def expr_property(expr, prop):
    if prop == 'prodcount':
        return prod_count(expr)
    elif prop == 'prodheight':
        return prod_height(expr)
    elif prop == 'argcount':
        if prod_node(expr):
            return len(prod_args(expr))
        else:
            return -1
    else:
        raise AZeeError('Unhandled property name: ' + prop)


def recursive_node_match(template, expr):
    res = []
    if root_match := template.templateMatch(expr):
        res.append(root_match)
    for parg in prod_args(expr):
        res.extend(recursive_node_match(template, parg))
    return res





################################
#                              #
#      MAIN TEST FUNCTION      #
#                              #
################################

def template_test(t_op, t_args, expr):
    result = None

    ##  BOOLEAN LOGIC
    if t_op in ['and', 'or', 'not']:
        if t_op == 'not' and len(t_args) != 1:
            raise AZeeError('Expected single argument test for "?not" template.')
        pretests = [arg.templateMatch(expr) for arg in t_args]
        if any(match.wildcards() for match in pretests if match):  # FIXME should be testing the templates, not the matches
            raise AZeeError('Cannot mix boolean template logic with named wildcards.') # NOTE this is also triggered by ?prodargcount test
        if t_op == 'and' and all(match for match in pretests) \
                or t_op == 'or' and any(match for match in pretests) \
                or t_op == 'not' and not pretests[0]:
            result = TemplateMatch(expr)

    elif t_op == 'suchthat':
        if len(t_args) < 2:
            raise AZeeError('Bad number of arguments for "?suchthat" template.')
        pretests = [arg.templateMatch(expr) for arg in t_args]
        last = pretests.pop()  # the only test with wilcards allowed
        if any(match.wildcards() for match in pretests if match): # FIXME should be testing the templates, not the matches
            raise AZeeError('Cannot use named wildcards in "?suchthat" except for last test.')
        if all(match for match in pretests) and last:
            result = TemplateMatch(expr, wilcards=last.wildcardDict())


    ##  TESTS ON PRODUCTION NODES
    elif t_op.startswith(':') or t_op.startswith('.'):
        if (expr.isOperator('apply') or t_op[0] == '.' and expr.isOperator('partialapply')) and (ref := expr.arg(0).reference()):
            if t_op[1:].startswith('*'):  # wildcard for name of applied ref
                if len(t_op[1:]) > 1:  # has a key to match (wildcard is not just '*' alone)
                    result = TemplateMatch(expr, {t_op[1:]: expr.arg(0)}) # saving reference AZexpr
                else:
                    result = TemplateMatch(expr)  # wildcard or condition on applied named ref below
            elif ref == t_op[1:]:  # matching explicitly named rule
                result = TemplateMatch(expr)

            if result and t_op[1:].startswith('*') and t_args:
                AZeeError('"?:/." should not be used with both a named wildcard and conditions on args.', t_args[0].srcPos()).stderrWarning()

            i = 0
            while result and i < len(t_args):
                trace('Arg. condition %s' % t_args[i], channel='template')
                try:
                    idx = next(j for j, arg in enumerate(expr.args()) if t_args[i].isValue(t=Type.STRING)
                            and arg.isValue(t=Type.STRING) and t_args[i] >> Type.STRING == arg >> Type.STRING)
                    if i + 1 < len(t_args) and not t_args[i + 1].isValue(t=Type.STRING):
                        if rec_res := t_args[i + 1].templateMatch(expr.arg(idx + 1)):
                            result.extendWith(rec_res)
                        else:
                            result = None
                        i += 2
                    else:  # no test for arg (implicit wildcard, only existence is tested)
                        i += 1
                except StopIteration:  # argument not applied in the tested expression
                    trace('not found in expr', channel='template')
                    result = None


    ##  any azop applied with all/any args verifying a condition
    elif t_op in ['anyarg', 'allargs']:
        if prod_node(expr):
            template = t_args[0]
            tests = [template.templateMatch(arg) for arg in prod_args(expr)]
            if any(match.wildcards() for match in tests if match): # FIXME should be testing the templates, not the matches
                raise AZeeError('Cannot mix "anyarg" or "allargs" with named wildcards.')
            if t_op == 'anyarg' and any(match for match in tests) \
                    or t_op == 'allargs' and all(match for match in tests):
                result = TemplateMatch(expr)

    # azop applied with test on number of arg's verifying a condition
    elif t_op.startswith('argcount'):
        if prod_node(expr):
            tests = [template_test('and', t_args, arg) for arg in prod_args(expr)]
            if int_comp_op_test(t_op, len([match for match in tests if match])):
                result = TemplateMatch(expr)

    # test on production category
    elif t_op.startswith('prodcat='):
        if rule := prod_node(expr):
            expected = t_op.split('=', maxsplit=1)[1]
            if expected == 'lexical':  # single hardcoded category for now; FUTURE linguist-defined rule categories
                if expr.possibleInfamousLexicalFallback():
                    result = TemplateMatch(expr)
                else: # was an infamous lexical fallback
                    arg_types, arg_defaults, azop_body = local_db_deref(rule).eval({}, dummy_posture) >> Type.AZOP
                    if all(arg in arg_defaults for arg in arg_types):
                        result = TemplateMatch(expr)
            else:
                raise AZeeError('Undefined rule category "%s"' % expected)

    # test on production height of AZee tree
    elif t_op.startswith('prodheight'):
        if int_comp_op_test(t_op, prod_height(expr)):
            result = TemplateMatch(expr)

    # test on production node count
    elif t_op.startswith('prodcount'):
        if int_comp_op_test(t_op, prod_count(expr)):
            result = TemplateMatch(expr)

    # OTHER TESTS
    # root or any offspring node verifies a condition
    elif t_op == 'contains':
        if expr_contains(expr, t_args[0]):
            result = TemplateMatch(expr)

    # expr type test
    elif t_op.startswith('type='):
        expected = TypeSpec.fromStr(t_op.split('=', maxsplit=1)[1])
        if not expr.typeSpec().acceptsAll() and expected.accepts(expr.typeSpec()):
            result = TemplateMatch(expr) # avoid evaluation if type check can give us expression type
        elif expected.acceptsOnlyBasic(expr.eval({}, dummy_posture).type):
            result = TemplateMatch(expr)

    # direct expression operator test
    elif t_op.startswith('isvalue'):
        if expr.isValue():
            result = TemplateMatch(expr)
    elif t_op.startswith('op='):
        if expr.isOperator(t_op.split('=', maxsplit=1)[1]):
            result = TemplateMatch(expr)
    else:
        raise AZeeError('Invalid template test: %s' % t_op)
    return result
