from math import fabs, sqrt, pi, sin, cos, asin, acos, atan2


class vec3:
    """
    vec3
    Class supporting the use of three-component vectors
    """

    @staticmethod
    def unit(axis: str):
        """
        :rtype: vec3
        """
        return {'x': vec3(1, 0, 0), 'y': vec3(0, 1, 0), 'z': vec3(0, 0, 1)}[axis.lower()]

    def __init__(self, x: float=0.0, y: float=0.0, z: float=0.0):
        """
        New vector from coordinates, default vector : (0, 0, 0)
        Params :
            - x : float
            - y : float
            - z : float

        :rtype: vec3
        """
        self.x = float(x)
        self.y = float(y)
        self.z = float(z)

    def __eq__(self, o) -> bool:
        """ Equality
        :type o: vec3
        """
        return (fabs(self.x - o.x) < 1e-6) and (fabs(self.y - o.y) < 1e-6) and (fabs(self.z - o.z) < 1e-6)

    def __neg__(self):
        return vec3(-self.x, -self.y, -self.z)

    def __add__(self, o):
        """ Addition
        :type o: vec3
        """
        return vec3(self.x + o.x, self.y + o.y, self.z + o.z)

    def __sub__(self, o):
        """ Subtraction
        :type o: vec3
        """
        return vec3(self.x - o.x, self.y - o.y, self.z - o.z)

    def __mul__(self, o):
        """
        Multiplication depending on the other operand :
            . with another vector : returns the scalar product
            . with a quaternion (class quat) : returns the quaternion product
            . with a scalar : vector scaling
        """
        if isinstance(o, vec3):
                return self.x * o.x + self.y * o.y + self.z * o.z
        elif isinstance(o, quat):
                return quat(xyz=self, w=0) * o
        else:
                return vec3(self.x * o, self.y * o, self.z * o)

    def __truediv__(self, o):
        """ Division by a scalar """
        return vec3(self.x / o, self.y / o, self.z / o)

    def length(self) -> float:
        """
        Length of the vector.
            - Returns a float
        """
        return sqrt(self.x * self.x + self.y * self.y + self.z * self.z)

    def normalize(self) -> None:
        """Normalization: resize this vector to length 1.0.
        """
        l = self.length()

        if fabs(l) > 1e-6:
            self.x /= l
            self.y /= l
            self.z /= l

    def normalized(self):
        """Returns a normalized version of this vector. The original is unchanged.
        :rtype: vec3
        """
        l = self.length()
        return vec3(x=self.x / l, y=self.y / l, z=self.z / l)

    def cross(self, o):
        """
        Cross product
            :rtype: vec3
        """
        x = self.y * o.z - self.z * o.y
        y = self.z * o.x - self.x * o.z
        z = self.x * o.y - self.y * o.x

        return vec3(x, y, z)

    def to_list(self) -> list:
        """
        Returning the vector under a list form
            - Returns a [float, float, float]
        """
        return [self.x, self.y, self.z]

    # def to_mat(self):
    # 		"""
    # 		Returning the vector under a numpy.matlib.mat form.
    # 			- Returns a numpy.matlib.mat
    # 		"""
    # 		return (matlib.mat([self.x, self.y, self.z]))

    def __getitem__(self, index: int):
        if index == 0:
            return self.x
        elif index == 1:
            return self.y
        elif index == 2:
            return self.z
        else:
            raise BaseException("Index out of range "+str(index))

    def __setitem__(self, index: int, value: float):
        if index == 0:
            self.x = value
        elif index == 1:
            self.y = value
        elif index == 2:
            self.z = value
        else:
            raise BaseException("Index out of range "+str(index))

    def __str__(self):
            """ Returns the vector as a string """
            return "<{0}, {1}, {2}>".format(self.x, self.y, self.z)

    def str_list(self):
            """ Returns the vector as a string list """
            return [str(self.x), str(self.y), str(self.z)]

    #  ADDED BY MICKY:
    def rot(self, quat_rot):
        """
        returns the point vec3 image of self after applying given quaternion-represented rotation
        """
        p2 = quat_rot * quat(xyz=self, w=0) * quat_rot.conj()
        return vec3(p2.x, p2.y, p2.z)


#
# =============================
#


class quat:
    """
    Class supporting the use of quaternions to represent rotations
    """

    @staticmethod
    def EulerRotation(x_rot: float, y_rot: float, z_rot: float, order: str):
        """
        :rtype: quat
        """
        q = quat()  # identity rotation
        for axis in order:
            rot_angle = {'x': x_rot, 'y': y_rot, 'z': z_rot}[axis]
            q = quat.rotation(vec3.unit(axis), rot_angle) * q
        return q

    @staticmethod
    def rotation(axis: vec3, angle: float):
        """
        returns a quat representing the specified rotation
        :rtype: quat
        """
        a2 = angle / 2
        return quat(axis * (sin(a2) / axis.length()), cos(a2))

    @staticmethod
    def identity():
        """
        :rtype: quat
        """
        return quat(xyz=vec3(0, 0, 0), w=1)

    #  REWRITTEN BY MICKY
    # ... and then rewritten by Fabrizio
    def __init__(self, xyz: vec3 = vec3(0, 0, 0), w: float = 1):
        """
        :param xyz: a vec3 instance
        :param w: a float
        """

        self.w = w
        self.x = xyz.x
        self.y = xyz.y
        self.z = xyz.z

        # if isinstance(arg1, vec3):  # WHAT! Why is this test failing after reloading the addon???!!!
        # 	self.w = 0.0 if arg2 == None else float(arg2)
        # 	self.x = float(arg1.x)
        # 	self.y = float(arg1.y)
        # 	self.z = float(arg1.z)
        # else:
        # 	self.w = float(arg1)
        # 	if arg2 is None:
        # 		self.x = self.y = self.z = 0.0
        # 	else:
        # 		self.x = float(arg2.x)
        # 		self.y = float(arg2.y)
        # 		self.z = float(arg2.z)

    def __eq__(self, o) -> bool:
        """ Equality. Also works with a vec3
        :type o: quat
        """
        if isinstance(o, vec3):
                return self.to_vector() == o

        return (fabs(self.x - o.x) < 1e-6) and (fabs(self.y - o.y) < 1e-6) and\
               (fabs(self.z - o.z) < 1e-6) and (fabs(self.w - o.w) < 1e-6)

    def __add__(self, o):
        """ Addition
        :rtype: quat
        :type o: quat
        """
        return quat(vec3(self.x + o.x, self.y + o.y, self.z + o.z), self.w + o.w)

    def __sub__(self, o):
        """ Subtraction
        :rtype: quat
        :type o: quat
        """
        return quat(vec3(self.x - o.x, self.y - o.y, self.z - o.z), self.w - o.w)

    def is_identity(self, tolerant=True) -> bool:
        if tolerant:
            return self == quat.identity()
        else:
            return self.w == 1 and self.x == 0 and self.y == 0 and self.z == 0

    def conj(self):
        """Conjugate quaternion (-x, -y, -z, w)
        :rtype: quat
        """
        return quat(vec3(-self.x, -self.y, -self.z), self.w)

    def __mul__(self, o):
            """ Multiplication, quat * quat or quat * vec3 """
            if isinstance(o, quat):
                    x1, y1, z1, w1 = self.x, self.y, self.z, self.w
                    x2, y2, z2, w2 = o.x, o.y, o.z, o.w

                    x = w1*x2 + x1*w2 + y1*z2 - z1*y2
                    y = w1*y2 - x1*z2 + y1*w2 + z1*x2
                    z = w1*z2 + x1*y2 - y1*x2 + z1*w2
                    w = w1*w2 - x1*x2 - y1*y2 - z1*z2

                    return quat(vec3(x, y, z), w)
            elif isinstance(o, vec3):
                    return self * quat(xyz=o, w=0)
            else:
                raise BaseException("Can not multiply a quat with an instance of "+str(type(o)))

    def eq_tolerance(self, o, tolerance) -> bool:
        return (fabs(self.x - o.x) < tolerance) and (fabs(self.y - o.y) < tolerance) and\
               (fabs(self.z - o.z) < tolerance) and (fabs(self.w - o.w) < tolerance)

    def magnitude(self) -> float:
        return sqrt(self.x * self.x + self.y * self.y + self.z * self.z + self.w * self.w)

    def normalize(self) -> None:
            """
            Normalization
                - Returns a quat of length 1
            """
            l = self.magnitude()  # sqrt(self.x * self.x + self.y * self.y + self.z * self.z + self.w * self.w)

            if fabs(l) > 1e-6:
                self.x /= l
                self.y /= l
                self.z /= l
                self.w /= l

    def to_vector(self) -> vec3:
            """
            Transforms to a vector.
                - Returns a vec3
            """
            return vec3(self.x, self.y, self.z)

    def __str__(self):
            """ Returns a string of the quaternion """
            return '[{0}, {1}, {2}, {3}]'.format(self.x, self.y, self.z, self.w)

    def axis_angle(self) -> (float, vec3):
            angle = 2.0 * acos(self.w)

            dif_angle = sin(angle / 2.0)

            if fabs(dif_angle) < 1e-5:
                    axis = vec3(1.0, 0.0, 0.0)
            else:
                    axis = vec3(self.x, self.y, self.z) / dif_angle

            return [angle, axis]

    def axis_angle_str(self):
            """ Returns a string of the axis/vector version of the quaternion """
            [angle, axis] = self.axis_angle()

            return '{0} {1} {2} {3}'.format(angle, axis.x, axis.y, axis.z)

    def to_euler(self):  # CHECKME are we not missing a vector order here?
            a = atan2(2*(self.w*self.x + self.y*self.z), 1 - 2*(self.x*self.x + self.y*self.y))
            b = asin(2*(self.w*self.y - self.z*self.x))
            c = atan2(2*(self.w*self.z + self.x*self.y), 1 - 2*(self.y*self.y + self.z*self.z))
            return a, b, c

    def dup(self):
        return quat(vec3(self.x, self.y, self.z), self.w)



#
# =====  GEOMETRIC PATHS  =====
#

class Path:
    # abstract method to override
    def length(self):
        return NotImplemented

    # abstract method to override
    def pointOnPath(self, t):
        return NotImplemented

    def startPoint(self):
        return self.pointOnPath(0)

    def endPoint(self):
        return self.pointOnPath(1)


class StraightPath(Path):
    def __init__(self, start, end):
        Path.__init__(self)
        self.start = start
        self.end = end

    def __str__(self):
        return 'straight from %s to %s' % (self.start, self.end)

    def length(self):
        return (self.end - self.start).length()

    def pointOnPath(self, t):
        return self.start + (self.end - self.start) * t



class ArcPath(Path):
    def __init__(self, centre, start, nrm_vect, rev_count):
        Path.__init__(self)
        self.centre = centre
        self.start = start
        self.nrm_vect = nrm_vect.normalized()
        self.rev_count = rev_count

    def __str__(self):
        return 'arc around %s (%f circle)' % (self.centre, self.rev_count)

    def length(self):
        return 2 * pi * (self.start - self.centre).length() * self.rev_count

    def pointOnPath(self, t):
        return (self.start - self.centre).rot(quat.rotation(self.nrm_vect, -2 * pi * self.rev_count * t)) + self.centre



class SinePath(Path):
    segments_per_wave = 16 # for approximating length

    def __init__(self, start, vect_to_end, nrm_vect, amplitude, wave_count, phase):
        Path.__init__(self)
        self.start = start
        self.dir_vect = vect_to_end.normalized() # FIXME a rotation is missing; this should be used somewhere
        self.nrm_vect = nrm_vect.normalized()
        self.linear_distance = vect_to_end.length()
        self.amplitude = amplitude
        self.wave_count = wave_count
        self.phase = phase
        self.sin2d = lambda t: self.amplitude * sin(2 * pi * t * self.wave_count + self.phase)
        self.rot3d = quat(vec3(-self.nrm_vect.y, self.nrm_vect.x, 0), w=self.nrm_vect.z) # FIXME this should include a rot turning X into dir_vect

    def __str__(self):
        return 'sine from %s (%f waves)' % (self.start, self.wave_count)

    def length(self): # this is approximate; based on SinePath.segments_per_wave
        nseg = max(1, int(SinePath.segments_per_wave * self.wave_count))
        return sum((self.pointOnPath((t + 1) / nseg) - self.pointOnPath(t / nseg)).length() for t in range(nseg))

    def pointOnPath(self, t):
        return self.start + vec3(t * self.linear_distance, self.sin2d(t), 0).rot(self.rot3d)



class CatmullRomPath(Path):
    # point at t=0~1 between Pi and Pi+1 is .5*tmat*mmat*pmat (CHECKME: I am interpreting .5 as the generic "tau" tension factor)
    mmat = [[0, 2, 0, 0], [-1, 0, 1, 0], [2, -5, 4, -1], [-1, 3, -3, 1]]

    def __init__(self, spec_points, tau=.5, resolution=10): # resolution is number of segments between consecutive spec points
        Path.__init__(self)
        self.spec_points = spec_points
        self._path_points = [spec_points[0]]
        control_points = [spec_points[0]] + spec_points + [spec_points[-1]] # end control points are added on either side
        for i in range(len(spec_points) - 1):
            pmat = [xyz_list(control_points[i + j]) for j in range(4)]
            mat2 = matrix_mult(CatmullRomPath.mmat, pmat)
            for j in range(1, resolution):
                t = j / resolution
                tmat = [[1, t, t*t, t*t*t]]
                mat3 = matrix_mult(tmat, mat2) # one-row matrix
                self._path_points.append(vec3(*mat3[0]) * tau)
            self._path_points.append(spec_points[i + 1])
        self._distances = [(self._path_points[i + 1] - self._path_points[i]).length() for i in range(len(self._path_points) - 1)]

    def __str__(self):
        return 'Catmull-Rom curve through %i specified points' % len(self.spec_points)

    def length(self): # this is approximate; based on small chunks following the init resolution
        return sum(self._distances)

    def pointOnPath(self, t):
        if 0 <= t <= 1:
            target = t * self.length()
            acc = 0
            for i, d in enumerate(self._distances): # length one fewer than self._path_points
                if acc + d >= target:
                    return StraightPath(self._path_points[i], self._path_points[i + 1]).pointOnPath((target - acc) / d)
                else:
                    acc += d
        elif t < 0:
            return self._path_points[0] + (self._path_points[1] - self._path_points[0]).normalized() * (self.length() * t)
        else: # t > 1
            return self._path_points[-1] + (self._path_points[-1] - self._path_points[-2]).normalized() * self.length() * (t - 1)


def matrix_mult(a, b):
    """
    Matrices are lists of row lists.
    """
    return [[sum(a * b for a, b in zip(a_row, b_col)) for b_col in zip(*b)] for a_row in a]


def xyz_list(v):
    return [v.x, v.y, v.z]
