
import re
from sys import stderr


class AZeeError(Exception):
    warning_types = WHITESPACE, INFAMOUS_LEXICAL_FALLBACK, UNBOUND_REF_DEPS, BINDING, LABELLED_ARG = range(5)
    silenced_warnings = set()
    wrn_txt_keys = dict(zip(['ws', 'ilf', 'deps', 'bind', 'arg'], warning_types, strict=True))

    def __init__(self, msg, src=None):
        self.msg = msg
        self.src_pos = src

    def position(self):
        return self.src_pos

    def setSrcPos(self, src_pos):
        self.src_pos = src_pos

    def descr(self, lvlPrefix=None):
        parts = []
        if lvlPrefix:
            parts.append(lvlPrefix)
        if self.src_pos:
            parts.append('[%s]' % self.src_pos)
        parts.append(self.msg)
        return ' '.join(parts)

    def stderrWarning(self, cat=None):
        if cat is None or cat not in AZeeError.silenced_warnings:
            print(self.descr(lvlPrefix='WARNING'), file=stderr)

    def stderrError(self):
        print(self.descr(lvlPrefix='ERROR'), file=stderr)




class Type:
    aztypes = NUM, STRING, LIST, POINT, VECT, PATH, \
            SIDE, BONE, POSTURE, SCORE, AZOP = range(11)

    @staticmethod
    def fromStr(s):
        try:
            return next(t for t in Type.aztypes if Type.toStr(t) == s)
        except StopIteration:
            raise AZeeError('Invalid type specification: %s' % s)

    @staticmethod
    def toStr(t):
        return {
            Type.NUM:  'NUM',   Type.SIDE:  'SIDE',  Type.POINT:   'POINT',
            Type.VECT: 'VECT',  Type.BONE:  'BONE',  Type.STRING:  'STRING',
            Type.AZOP: 'AZOP',  Type.LIST:  'LIST',  Type.POSTURE: 'POSTURE',
            Type.PATH: 'PATH',  Type.SCORE: 'SCORE'
        }[t]



# the following is assumed to be run on strings stripped of all spaces
azop_type_spec_re = re.compile(r'(AZOP|(AZOP)?\((?P<args>(\??\w+(:\w+)?(,\??\w+(:\w+)?)*)?)\))(->(?P<res>\w+))?')

class TypeSpec:
    def __init__(self, t): # arg is basic aztype to enforce, or None if anything should match
        self._base = t
        self._azop_args = None # if AZOP and arg list known: dict(str -> "arg is optional" bool, TypeSpec)
        self._azop_res = None  # if AZOP: TypeSpec
        if t == Type.AZOP:
            self._azop_res = TypeSpec(None) # won't recurse

    @staticmethod
    def fromValue(val):
        res = TypeSpec(val.type)
        if val.type == Type.AZOP:
            arg_types, arg_defaults, body = val.value
            res._azop_args = {p: (p in arg_defaults, t) for p, t in arg_types.items()}
            res._azop_res = body.typeSpec()
        return res

    @staticmethod
    def fromStr(s):
        """
        Typically used to specify AZOP arguments that have no default value
        """
        if match := azop_type_spec_re.fullmatch(s.replace(' ', '')):
            res = TypeSpec(Type.AZOP)
            if match['args'] is not None: # got argument list spec; can be empty string
                res._azop_args = {}
                for argspec in match['args'].split(','):
                    noqspec = argspec.lstrip('?')
                    if ':' in noqspec:
                        arg, t = noqspec.split(':', maxsplit=1)
                        tspec = TypeSpec.fromStr(t)
                    else: # no type constraint given for this named argument
                        arg = noqspec
                        tspec = TypeSpec(None)
                    res._azop_args[arg] = argspec.startswith('?'), tspec
            if resstr := match['res']: # got result spec; "res" group cannot be empty string
                res._azop_res = TypeSpec.fromStr(resstr)
            return res
        else: # should be just a basic aztype string
            return TypeSpec(Type.fromStr(s))

    def acceptsAll(self):
        return self._base is None

    def acceptsOnlyBasic(self, t):
        return self._base == t

    def accepts(self, other):
        if self._base == Type.AZOP and other._base == Type.AZOP:
            if self._azop_args is not None and other._azop_args is not None: # check if args are compatible
                for myarg, (myopt, mytyp) in self._azop_args.items():
                    try:
                        opt, typ = other._azop_args[myarg]
                        if myopt and not opt or not mytyp.accepts(typ):
                            return False
                    except KeyError: # other azop does not have such argument
                        return False
            if self._azop_res is not None and other._azop_res is not None: # check if result type is compatible
                if not self._azop_res.accepts(other._azop_res):
                    return False
            return True
        else:
            return self._base is None or other._base is None or self._base == other._base

    def toStr(self):
        res = 'any' if self._base is None else Type.toStr(self._base)
        if self._base == Type.AZOP:
            if self._azop_args is not None:
                res += ' (%s)' % ', '.join(_azopArgSpec_toStr(opt, arg, tspec) for arg, (opt, tspec) in self._azop_args.items())
            if not self._azop_res.acceptsAll():
                res += ' -> %s' % self._azop_res.toStr()
        return res


def _azopArgSpec_toStr(opt_flag, arg_name, type_spec):
    res = '?' if opt_flag else ''
    res += arg_name
    if not type_spec.acceptsAll():
        res += ': ' + type_spec.toStr()
    return res



class AZval:
    def __init__(self, typ, val):
        self.type = typ
        self.value = val
        self._src_expr = None # NOTE: score values contain their own copy of the source expression

    def __eq__(self, other): # CAUTION only use with what can be AZee input (list, score, etc. eval results won't compare easily)
        return self.type == other.type and self.value == other.value

    def __rshift__(self, wanted):
        if self.type == wanted:
            return self.value
        else:
            src_pos = None if self._src_expr is None else self._src_expr.srcPos()
            raise AZeeError('Expected %s; got %s.' % (Type.toStr(wanted), Type.toStr(self.type)), src=src_pos)

    def __str__(self):
        """
        parsable one-liner if possible, otherwise a one-line description of the value
        """
        if self.type == Type.STRING:
            return '\'' + self.value
        if self.type == Type.NUM:
            return '%g' % self.value
        if self.type == Type.VECT:
            return '<%g, %g, %g>' % (self.value.x, self.value.y, self.value.z)
        if self.type == Type.POINT:
            return '[%g, %g, %g]' % (self.value.x, self.value.y, self.value.z)
        # CAUTION: the following cases do reflect full spec of the value, but do not parse as such in input expressions
        if self.type == Type.SIDE:
            return 'LR'[self.value]
        if self.type == Type.BONE:
            return '!' + self.value
        # CAUTION: the following cases only provide a summary of the value, not a full spec
        if self.type == Type.AZOP:
            arg_types, arg_defaults, azop_body = self.value
            return '(%s) -> %s' % (', '.join('?' + arg if arg in arg_defaults else arg for arg in arg_types), azop_body.pipedStr())
        if self.type == Type.POSTURE:
            return ', '.join('%s (%i)' % (t, sum(cstr[0] == t for cstr in self.value)) for t in set(cstr[0] for cstr in self.value))
        if self.type == Type.LIST:
            return ', '.join(e.typeSpec().toStr() for e in self.value)
        if self.type == Type.PATH:
            return 'from %s to %s; length %f' % (AZval(Type.POINT, self.value.startPoint()), AZval(Type.POINT, self.value.endPoint()), self.value.length())
        if self.type == Type.SCORE:
            return '%i posture/hold blocks; duration hint %.2f' % (self.value.blockCount(), self.value.flattened().duration())
        raise ValueError('Unknown type format: %s' % self.type)

    def isType(self, t):
        return self.type == t

    def setSourceExpr(self, expr):
        self._src_expr = expr
        if self.isType(Type.SCORE):
            self.value.setSourceAZexpr(expr)
